from jaqpotpy.models.base_classes import Model
from jaqpotpy.doa.doa import DOA
from typing import Any, Union, Dict, Optional
from jaqpotpy.datasets import MolecularDataset
from jaqpotpy.datasets.material_datasets import CompositionDataset, StructureDataset
from jaqpotpy.models import Evaluator, Preprocesses, MolecularModel, MaterialModel
import sklearn
from jaqpotpy.cfg import config
import jaqpotpy
from skl2onnx import convert_sklearn
from skl2onnx.common.data_types import FloatTensorType
import onnxruntime as rt
import numpy as np


class MolecularSKLearn(Model):

    def __init__(self, dataset: MolecularDataset, doa: DOA, model: Any
                 , eval: Evaluator = None, preprocess: Preprocesses = None):
        # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
        self.dataset = dataset
        self.model = model
        self.doa = doa
        self.doa_m = None
        self.external = None
        self.evaluator: Evaluator = eval
        self.preprocess: Preprocesses = preprocess
        self.trained_model = None
        self._initial_types = None

    def __call__(self, smiles):
        self

    def fit(self):
        model = MolecularModel()
        if self.dataset.df is not None:
            pass
        else:
            self.dataset.create()
        # if not self.dataset.df.empy:
        #     self.dataset.create()
        # else:
        #     pass
        if self.doa:
            if self.doa.__name__ == 'SmilesLeverage':
                self.doa_m = self.doa.fit(self.dataset.smiles)
            else:
                self.doa_m = self.doa.fit(X=self.dataset.__get_X__())
        X = self.dataset.__get_X__()
        y = self.dataset.__get_Y__()
        if self.preprocess:
            pre_keys = self.preprocess.classes.keys()
            preprocess_names = []
            preprocess_classes = []
            for pre_key in pre_keys:
                pre_function = self.preprocess.classes.get(pre_key)
                X = pre_function.fit_transform(X)
                self.preprocess.register_fitted_class(pre_key, pre_function)
                preprocess_names.append(pre_key)
                preprocess_classes.append(pre_function)
            model.preprocessing = preprocess_classes
            model.preprocessor_names = preprocess_names

            pre_y_keys = self.preprocess.classes_y.keys()
            preprocess_names_y = []
            preprocess_classes_y = []
            for pre_y_key in pre_y_keys:
                pre_y_function = self.preprocess.classes_y.get(pre_y_key)
                y = pre_y_function.fit_transform(y)
                self.preprocess.register_fitted_class_y(pre_y_key, pre_y_function)
                preprocess_names_y.append(pre_y_key)
                preprocess_classes_y.append(pre_y_function)
            model.preprocessing_y = preprocess_classes_y
            model.preprocessor_y_names = preprocess_names_y

        X = X.astype(float).copy()
        self.trained_model = self.model.fit(X, y)
        if type(self.dataset.featurizer).__name__ == "RDKitDescriptors":
            model.descriptors = "RDKitDescriptors"
        elif type(self.dataset.featurizer).__name__ == "MACCSKeysFingerprint":
            model.descriptors = "MACCSKeysFingerprint"
        else:
            model.descriptors = self.dataset.featurizer
        model.doa = self.doa
        model.model = self.trained_model
        model.X = self.dataset.X
        model.Y = self.dataset.y
        model.library = ['sklearn']
        model.version = [sklearn.__version__]
        model.jaqpotpy_version = jaqpotpy.__version__
        model.jaqpotpy_docker = config.jaqpotpy_docker
        model.external_feats = self.dataset.external
        model.inference_model = self.__convert_to_onnx__(X.shape[1])
        self.inference_model = model.inference_model
        if self.evaluator:
            self.__eval__()
        return model

    def __convert_to_onnx__(self, num_columns):

        name = self.model.__class__.__name__ + "_ONNX"
        initial_types = [('float_input',FloatTensorType([None, num_columns]))]
        res = convert_sklearn(self.model, initial_types=initial_types, name=name, target_opset=None,
                              options=None, black_op=None, white_op=None, final_types=None,
                              verbose=0)
        return res

    def __predict__(self, X):
        pre_keys = self.preprocess.fitted_classes.keys()
        for pre_key in pre_keys:
            pre_function = self.preprocess.fitted_classes.get(pre_key)
            X = pre_function.transform(X)
        data = self.dataset.featurizer.featurize(X)
        sess = rt.InferenceSession(self.model.inference_model.SerializeToString())
        input_name = sess.get_inputs()[0].name
        pred_onx = sess.run(None, {input_name: data.astype(np.float32)})
        return pred_onx[0].flatten()

    def __eval__(self):
        if self.evaluator.dataset.df is not None:
            pass
        else:
            self.evaluator.dataset.create()
        X = self.evaluator.dataset.__get_X__()
        if self.preprocess:
            pre_keys = self.preprocess.classes.keys()
            for pre_key in pre_keys:
                pre_function = self.preprocess.fitted_classes.get(pre_key)
                X = pre_function.transform(X)
        sess = rt.InferenceSession(self.inference_model.SerializeToString())
        input_name = sess.get_inputs()[0].name
        X = np.array(X.astype(float).copy())
        preds = sess.run(None, {input_name: X.astype(np.float32)})
        preds = preds[0].flatten()
        # preds = self.trained_model.predict(X)
        preds_t = []
        for p in preds:
            try:
                if self.preprocessing_y:
                    for f in self.preprocessing_y:
                        p = f.inverse_transform(p.reshape(1, -1))
                        preds_t.append(p)
            except AttributeError as e:
                pass
            preds_t.append(p)
        eval_keys = self.evaluator.functions.keys()
        for eval_key in eval_keys:
            eval_function = self.evaluator.functions.get(eval_key)
            try:
                if self.preprocessing_y:
                    for f in self.preprocessing_y:
                        truth = f.inverse_transform(self.evaluator.dataset.__get_Y__())
                    print(eval_key + ": " + str(eval_function(truth, preds_t)))
            except AttributeError as e:
                print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds_t)))
                pass
            # print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds_t)))
            #print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds)))



class MaterialSKLearn(Model):

    def __init__(self, dataset: Union[StructureDataset, CompositionDataset], doa: DOA, model: Any, fillna: Optional[Union[Dict, float, int]] = None
                 , eval: Evaluator = None, preprocess: Preprocesses = None):
        # super(InMemMolModel, self).__init__(dataset=dataset, doa=doa, model=model)
        self.dataset = dataset
        self.model = model
        self.doa = doa
        self.doa_m = None
        self.external = None
        self.evaluator: Evaluator = eval
        self.preprocess: Preprocesses = preprocess
        self.trained_model = None
        self.fillna = fillna
        # self.trained_model = None

    def __call__(self, smiles):
        self

    def fit(self):
        model = MaterialModel()
        if self.dataset.df is not None:
            pass
        else:
            self.dataset.create()

        self.dataset.df.fillna(self.fillna, inplace=True)

        if self.doa:
            self.doa_m = self.doa.fit(X=self.dataset.__get_X__())

        X = self.dataset.__get_X__()
        y = self.dataset.__get_Y__()
        if self.preprocess:
            pre_keys = self.preprocess.classes.keys()
            preprocess_names = []
            preprocess_classes = []
            for pre_key in pre_keys:
                pre_function = self.preprocess.classes.get(pre_key)
                X = pre_function.fit_transform(X)
                self.preprocess.register_fitted_class(pre_key, pre_function)
                preprocess_names.append(pre_key)
                preprocess_classes.append(pre_function)
            model.preprocessing = preprocess_classes
            model.preprocessor_names = preprocess_names

            pre_y_keys = self.preprocess.classes_y.keys()
            preprocess_names_y = []
            preprocess_classes_y = []
            for pre_y_key in pre_y_keys:
                pre_y_function = self.preprocess.classes_y.get(pre_y_key)
                y = pre_y_function.fit_transform(y)
                self.preprocess.register_fitted_class_y(pre_y_key, pre_y_function)
                preprocess_names_y.append(pre_y_key)
                preprocess_classes_y.append(pre_y_function)
            model.preprocessing_y = preprocess_classes_y
            model.preprocessor_y_names = preprocess_names_y

        self.trained_model = self.model.fit(X, y)
        model.descriptors = self.dataset.featurizer
        model.doa = self.doa
        model.model = self.trained_model
        model.X = self.dataset.X
        model.Y = self.dataset.y
        model.library = ['sklearn']
        model.version = [sklearn.__version__]
        model.jaqpotpy_version = jaqpotpy.__version__
        model.external_feats = self.dataset.external
        if self.evaluator:
            self.__eval__()
        return model

    def __predict__(self, X):
        pre_keys = self.preprocess.fitted_classes.keys()
        for pre_key in pre_keys:
            pre_function = self.preprocess.fitted_classes.get(pre_key)
            X = pre_function.transform(X)
        data = self.dataset.featurizer.featurize_dataframe(X)
        data.df.fillna(self.fillna, inplace=True)
        data = data[self.dataset.X].to_numpy()
        return self.model.predict(data)

    def __eval__(self):
        if self.evaluator.dataset.df is not None:
            pass
        else:
            self.evaluator.dataset.create()
        self.evaluator.dataset.df.fillna(self.fillna, inplace=True)
        X = self.evaluator.dataset.__get_X__()
        if self.preprocess:
            pre_keys = self.preprocess.classes.keys()
            for pre_key in pre_keys:
                pre_function = self.preprocess.fitted_classes.get(pre_key)
                X = pre_function.transform(X)
        preds = self.trained_model.predict(X)
        eval_keys = self.evaluator.functions.keys()
        for eval_key in eval_keys:
            eval_function = self.evaluator.functions.get(eval_key)
            print(eval_key + ": " + str(eval_function(self.evaluator.dataset.__get_Y__(), preds)))


import numpy as np

from jaqpotpy.utils.types import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer


class PubChemFingerprint(MolecularFeaturizer):
    """PubChem Fingerprint.
  The PubChem fingerprint is a 881 bit structural key,
  which is used by PubChem for similarity searching.
  Please confirm the details in [1]_.
  References
  ----------
  .. [1] ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.pdf
  Note
  -----
  This class requires RDKit and PubChemPy to be installed.
  PubChemPy use REST API to get the fingerprint, so you need the internet access.
  Examples
  --------
  >>> import jaqpotpy as jt
  >>> smiles = ['CCC']
  >>> featurizer = jt.descriptors.PubChemFingerprint()
  >>> features = featurizer.featurize(smiles)
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (881,)
  """

    @property
    def __name__(self):
        return 'PubChemFingerprint'

    def __init__(self):
        """Initialize this featurizer."""
        try:
            from rdkit import Chem  # noqa
            import pubchempy as pcp  # noqa
        except ModuleNotFoundError:
            raise ImportError("This class requires PubChemPy to be installed.")

        self.get_pubchem_compounds = pcp.get_compounds

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
    Calculate PubChem fingerprint.
    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit Mol object
    Returns
    -------
    np.ndarray
      1D array of RDKit descriptors for `mol`. The length is 881.
    """
        try:
            from rdkit import Chem
            import pubchempy as pcp
        except ModuleNotFoundError:
            raise ImportError("This class requires PubChemPy to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        smiles = Chem.MolToSmiles(datapoint)
        pubchem_compound = pcp.get_compounds(smiles, 'smiles')[0]
        feature = [int(bit) for bit in pubchem_compound.cactvs_fingerprint]
        return np.asarray(feature)
    
    def _get_column_names(self, **kwargs) -> list:
        descriptors_length = 881
        descriptors = []
        for i in range(descriptors_length):
            descriptors.append("f" + str(i))
        return descriptors

    def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        return self._featurize(datapoint, **kwargs)




from jaqpotpy.docking.pose_generation import PoseGenerator
from jaqpotpy.docking.pose_generation import VinaPoseGenerator
from jaqpotpy.docking.pose_generation import GninaPoseGenerator
from jaqpotpy.docking.docking import Docker
from jaqpotpy.docking.binding_pocket import ConvexHullPocketFinder


from jaqpotpy.helpers.logging import init_logger
try:
    from jaqpotpy.descriptors.material.geometrical import GeomDescriptors
    from jaqpotpy.descriptors.material.element_property_fingerprint import ElementPropertyFingerprint
    # from jaqpotpy.descriptors.material.cgcnn.graph_data import GraphData
    from jaqpotpy.descriptors.material.cgcnn import CrystalGraphCNN
    from jaqpotpy.descriptors.material.elemnet import ElementNet
    from jaqpotpy.descriptors.material.sine_coulomb_matrix import SineCoulombMatrix
except ModuleNotFoundError as e:
    logger = init_logger(__name__, testing_mode=False, output_log_file=False)
    logger.log(str(e))



import contextlib
import io
import sys
import unittest
import os
from datetime import datetime

@contextlib.contextmanager
def err_to(file):
    old_err = sys.stderr
    sys.stderr = file
    yield
    sys.stderr = old_err


if __name__ == '__main__':
    os.chdir('./run_tests')

    loader = unittest.TestLoader()
    start_dir = '../'
    suite = loader.discover(start_dir)

    result = io.StringIO()
    with err_to(result):
        runner = unittest.TextTestRunner()
        runner.run(suite)
        # unittest.main(exit=False)
    result.seek(0)

    #delete created files
    files = [item for item in os.listdir() if item != 'run_tests.py' and \
             item != '__init__.py' and \
             item != '__pycache__' and \
             'test_logs' not in item]

    for item in files:
        os.remove(item)

    timestamp = datetime.now()
    with open(f'./test_logs-{timestamp}'.replace(':','.').replace(' ','-'),'w') as logs:
        logs.write(result.read())


# from jaqpotpy.simulation_kit.lattice.base_classes import BravaisLattice
# from jaqpotpy.simulation_kit.lattice.cubic.cubic_lattice import SimpleCubic, FaceCenteredCubic, BodyCenteredCubic
# from jaqpotpy.simulation_kit.lattice.orthorhombic.orthorhombic_lattice import SimpleOrthorhombic, BaseCenteredOrthorhombic, BodyCenteredOrthorhombic, FaceCenteredOrthorhombic
# from jaqpotpy.simulation_kit.lattice.tetragonal.tetragonal_lattice import SimpleTetragonal, CenteredTetragonal
# from jaqpotpy.simulation_kit.lattice.hexagonal.hexagonal_lattice import SimpleHexagonal
# from jaqpotpy.simulation_kit.lattice.monoclinic.monoclinic_lattice import BaseCenteredMonoclinic, SimpleMonoclinic
# from jaqpotpy.simulation_kit.lattice.triclinic.triclinic_lattice import SimpleTriclinic
# from jaqpotpy.simulation_kit.periodic_table.base_classes import PeriodicTable
# from jaqpotpy.simulation_kit.periodic_table.element import Element
# from jaqpotpy.simulation_kit.composition.composition import Composition
# from jaqpotpy.simulation_kit.sites.sites import Site, PeriodicSite
# from jaqpotpy.simulation_kit.sites.neighbors import Neighbor, PeriodicNeighbor
# from jaqpotpy.simulation_kit.bonds.bonds import Bond


import json
import pandas as pd

def decode_predicted(dataset):
    feat_info = dataset['features']
    predicted = {}
    predicted_f = []
    for feat in feat_info:
        name = feat['name']
        key = feat['key']
        try:
            category = feat['category']
            if (category == 'PREDICTED') and (feat['name'] not in predicted_f):
                predicted_f.append(feat['name'])
        except KeyError:
            dataentries = {}
            for dataEntry in dataset['dataEntry']:
                dataentries[dataEntry['entryId']['name']] = dataEntry['values'][key]
            predicted[name] = dataentries
        dataentries = {}
        for dataEntry in dataset['dataEntry']:
            dataentries[dataEntry['entryId']['name']] = dataEntry['values'][key]
        predicted[name] = dataentries
    json_to_df = json.dumps(predicted)
    df = pd.read_json(json_to_df)
    return df, predicted_f

            # for val in dataEntry['values']:
#         for dataEntry in dataset['dataEntry']:
#             dataEntryToInsert = []
#             for key in shorted:
#                 dataEntryToInsert.append(dataEntry['values'][key])
#             dataEntryAll.append(dataEntryToInsert)
    #     for key in independentFeatures:
    #         if actual == independentFeatures[key]:
    #             for feature in pred_request.dataset['features']:
    #                 if feature['name'] == actual:
    #                     shorted.append(feature['key'])
    #             # shorted.append(key)
    # dataEntryAll = []
    # for dataEntry in pred_request.dataset['dataEntry']:
    #     dataEntryToInsert = []
    #     for key in shorted:
    #         dataEntryToInsert.append(dataEntry['values'][key])
    #     dataEntryAll.append(dataEntryToInsert)
    # return dataEntryAll


"""
Tests for binding pocket detection.
"""
import os
import logging
import unittest
import numpy as np

import jaqpotpy as jp
import jaqpotpy.docking as dock
from jaqpotpy.utils import rdkit_utils
# from jaqpotpy.docking.utils import coordinate_box_utils as box_utils

import jaqpotpy.docking.utils as box_utils

logger = logging.getLogger(__name__)


class TestBindingPocket(unittest.TestCase):
  """
  Does sanity checks on binding pocket generation.
  """

  # def test_convex_init(self):
  #   """Tests that ConvexHullPocketFinder can be initialized."""
  #   dc.dock.ConvexHullPocketFinder()

  def test_get_face_boxes_for_protein(self):
    """Tests that binding pockets are detected."""
    current_dir = os.path.dirname(os.path.realpath(__file__))
    protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    coords = rdkit_utils.load_molecule(protein_file)[0]

    boxes = box_utils.get_face_boxes(coords)
    assert isinstance(boxes, list)
    # Pocket is of form ((x_min, x_max), (y_min, y_max), (z_min, z_max))
    for pocket in boxes:
      assert isinstance(pocket, box_utils.CoordinateBox)

  def test_convex_find_pockets(self):
    """Test that some pockets are filtered out."""
    current_dir = os.path.dirname(os.path.realpath(__file__))
    protein_file = os.path.join(current_dir, "1jld_protein.pdb")

    finder = jp.docking.ConvexHullPocketFinder()
    all_pockets = finder.find_all_pockets(protein_file)
    pockets = finder.find_pockets(protein_file)
    # Test that every atom in pocket maps exists
    for pocket in pockets:
      assert isinstance(pocket, box_utils.CoordinateBox)

    assert len(pockets) < len(all_pockets)

  def test_extract_active_site(self):
    """Test that computed pockets have strong overlap with true binding pocket."""
    current_dir = os.path.dirname(os.path.realpath(__file__))
    protein_file = os.path.join(current_dir, "1jld_protein.pdb")
    ligand_file = os.path.join(current_dir, "1jld_ligand.sdf")

    active_site_box, active_site_coords = dock.binding_pocket.extract_active_site(protein_file, ligand_file)

    print(active_site_box)
    print(active_site_coords)

    assert isinstance(active_site_coords, np.ndarray)


from jaqpotpy.cfg import config
from jaqpotpy.datasets import SmilesDataset
from jaqpotpy.descriptors.molecular import MolGanFeaturizer, GraphMatrix
import torch
from torch_geometric.nn import GCNConv
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool


import pandas as pd
import numpy as np

import tensorflow as tf
from tensorflow import keras
from keras import layers

import matplotlib.pyplot as plt
from rdkit import Chem, RDLogger
from rdkit.Chem import BondType
from rdkit.Chem.Draw import MolsToGridImage

RDLogger.DisableLog("rdApp.*")

csv_path = keras.utils.get_file(
    "./250k_rndm_zinc_drugs_clean_3.csv",
    "https://raw.githubusercontent.com/aspuru-guzik-group/chemical_vae/master/models/zinc_properties/250k_rndm_zinc_drugs_clean_3.csv",
)

df = pd.read_csv("/Users/pantelispanka/.keras/datasets/250k_rndm_zinc_drugs_clean_3.csv")
df["smiles"] = df["smiles"].apply(lambda s: s.replace("\n", ""))
print(df.head())

smiles = df['smiles'].to_list()[0:1000]

feat = MolGanFeaturizer(max_atom_count=40)

dataset = SmilesDataset(smiles=smiles, featurizer=feat, task="generation")
dataset = dataset.create()
gm: GraphMatrix = dataset.df['MolGanGraphs'][0]


class Generator(torch.nn.Module):
    def __init__(self, in_channels
                 , num_layers, hidden_channels, out_channels
                 , activation: torch.nn.Module = torch.nn.ReLU(), dropout=0, norm: torch.nn.Module = None, act_first=True):
        super(Generator, self).__init__()
        torch.manual_seed(config.global_seed)

        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm

        self.layers = torch.nn.ModuleList([GCNConv(in_channels, hidden_channels)])
        for i in range(num_layers - 1):
            self.layers.append(GCNConv(hidden_channels, hidden_channels))
        self.out = torch.nn.Linear(hidden_channels, out_channels)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        for layer in self.layers:
            x = layer(x, edge_index)
            if self.norm and self.act_first is True:
                x = self.act(x)
                x = self.norm(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            elif self.norm and self.act_first is False:
                x = self.norm(x)
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            else:
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = global_mean_pool(x, data.batch)
        x = self.out(x)
        return x


class Discriminator(torch.nn.Module):
    def __init__(self, in_channels
                 , num_layers, hidden_channels, out_channels
                 , activation: torch.nn.Module = torch.nn.ReLU(), dropout=0, norm: torch.nn.Module = None, act_first=True):
        super(Discriminator, self).__init__()
        torch.manual_seed(config.global_seed)
        self.act = activation
        self.dropout = dropout
        self.act_first = act_first
        self.norm = norm
        self.layers = torch.nn.ModuleList([GCNConv(in_channels, hidden_channels)])
        for i in range(num_layers - 1):
            self.layers.append(GCNConv(hidden_channels, hidden_channels))
        self.out = torch.nn.Linear(hidden_channels, out_channels)

    def forward(self, data):
        x, edge_index = data.x, data.edge_index
        for layer in self.layers:
            x = layer(x, edge_index)
            if self.norm and self.act_first is True:
                x = self.act(x)
                x = self.norm(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            elif self.norm and self.act_first is False:
                x = self.norm(x)
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            else:
                x = self.act(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        x = global_mean_pool(x, data.batch)
        x = self.out(x)
        return x



from jaqpotpy.models.evaluator import Evaluator
from jaqpotpy.models.preprocessing import Preprocesses
from jaqpotpy.models.base_classes import MolecularModel, Model, MaterialModel
from jaqpotpy.models.sklearn import MolecularSKLearn, MaterialSKLearn
try:
    from jaqpotpy.models.torch_geometric import MolecularTorchGeometric
    from jaqpotpy.models.torch import MolecularTorch
    from jaqpotpy.models.torch_models.torch_geometric import *
    from jaqpotpy.models.torch import MolecularTorch, MaterialTorch
except ModuleNotFoundError:
    MolecularTorchGeometric = MolecularTorch = GCN = MolecularTorch = MaterialTorch = None
    pass


import math
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module


class GraphConvolutionLayer(Module):
    def __init__(self, in_features, u, activation, edge_type_num, dropout_rate=0.):
        super(GraphConvolutionLayer, self).__init__()
        self.edge_type_num = edge_type_num
        self.u = u
        self.adj_list = nn.ModuleList()
        for _ in range(self.edge_type_num):
            self.adj_list.append(nn.Linear(in_features, u))
        self.linear_2 = nn.Linear(in_features, u)
        self.activation = activation
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, n_tensor, adj_tensor, h_tensor=None):
        if h_tensor is not None:
            annotations = torch.cat((n_tensor, h_tensor), -1)
        else:
            annotations = n_tensor

        output = torch.stack([self.adj_list[i](annotations) for i in range(self.edge_type_num)], 1)
        output = torch.matmul(adj_tensor, output)
        out_sum = torch.sum(output, 1)
        out_linear_2 = self.linear_2(annotations)
        output = out_sum + out_linear_2
        output = self.activation(output) if self.activation is not None else output
        output = self.dropout(output)
        return output


class MultiGraphConvolutionLayers(Module):
    def __init__(self, in_features, units, activation, edge_type_num, with_features=False, f=0, dropout_rate=0.):
        super(MultiGraphConvolutionLayers, self).__init__()
        self.conv_nets = nn.ModuleList()
        self.units = units
        in_units = []
        if with_features:
            for i in range(len(self.units)):
                in_units = list([x + in_features for x in self.units])
            for u0, u1 in zip([in_features+f] + in_units[:-1], self.units):
                self.conv_nets.append(GraphConvolutionLayer(u0, u1, activation, edge_type_num, dropout_rate))
        else:
            for i in range(len(self.units)):
                in_units = list([x + in_features for x in self.units])
            for u0, u1 in zip([in_features] + in_units[:-1], self.units):
                self.conv_nets.append(GraphConvolutionLayer(u0, u1, activation, edge_type_num, dropout_rate))

    def forward(self, n_tensor, adj_tensor, h_tensor=None):
        hidden_tensor = h_tensor
        for conv_idx in range(len(self.units)):
            hidden_tensor = self.conv_nets[conv_idx](n_tensor, adj_tensor, hidden_tensor)
        return hidden_tensor


class GraphConvolution(Module):
    def __init__(self, in_features, graph_conv_units, edge_type_num, with_features=False, f_dim=0, dropout_rate=0.):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.graph_conv_units = graph_conv_units
        self.activation_f = torch.nn.Tanh()
        self.multi_graph_convolution_layers = \
            MultiGraphConvolutionLayers(in_features, self.graph_conv_units, self.activation_f, edge_type_num,
                                        with_features, f_dim, dropout_rate)

    def forward(self, n_tensor, adj_tensor, h_tensor=None):
        output = self.multi_graph_convolution_layers(n_tensor, adj_tensor, h_tensor)
        return output


class GraphConvolution2(Module):
    def __init__(self, in_features, out_feature_list, b_dim, dropout):
        super(GraphConvolution2, self).__init__()
        self.in_features = in_features
        self.out_feature_list = out_feature_list

        self.linear1 = nn.Linear(in_features, out_feature_list[0])
        self.linear2 = nn.Linear(out_feature_list[0], out_feature_list[1])

        self.dropout = nn.Dropout(dropout)

    def forward(self, inputs, adj, activation=None):
        # input : 16x9x9
        # adj : 16x4x9x9

        hidden = torch.stack([self.linear1(inputs) for _ in range(adj.size(1))], 1)
        hidden = torch.einsum('bijk,bikl->bijl', (adj, hidden))
        hidden = torch.sum(hidden, 1) + self.linear1(inputs)
        hidden = activation(hidden) if activation is not None else hidden
        hidden = self.dropout(hidden)

        output = torch.stack([self.linear2(hidden) for _ in range(adj.size(1))], 1)
        output = torch.einsum('bijk,bikl->bijl', (adj, output))
        output = torch.sum(output, 1) + self.linear2(hidden)
        output = activation(output) if activation is not None else output
        output = self.dropout(output)

        return output


class GraphAggregation(Module):
    def __init__(self, in_features, aux_units, activation, with_features=False, f_dim=0,
                 dropout_rate=0.):
        super(GraphAggregation, self).__init__()
        self.with_features = with_features
        self.activation = activation
        if self.with_features:
            self.i = nn.Sequential(nn.Linear(in_features+f_dim, aux_units),
                                   nn.Sigmoid())
            j_layers = [nn.Linear(in_features+f_dim, aux_units)]
            if self.activation is not None:
                j_layers.append(self.activation)
            self.j = nn.Sequential(*j_layers)
        else:
            self.i = nn.Sequential(nn.Linear(in_features, aux_units),
                                   nn.Sigmoid())
            j_layers = [nn.Linear(in_features, aux_units)]
            if self.activation is not None:
                j_layers.append(self.activation)
            self.j = nn.Sequential(*j_layers)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, n_tensor, out_tensor, h_tensor=None):
        if h_tensor is not None:
            annotations = torch.cat((out_tensor, h_tensor, n_tensor), -1)
        else:
            annotations = torch.cat((out_tensor, n_tensor), -1)
        # The i here seems to be an attention.
        i = self.i(annotations)
        j = self.j(annotations)
        output = torch.sum(torch.mul(i, j), 1)
        if self.activation is not None:
            output = self.activation(output)
        output = self.dropout(output)

        return output


class GraphAggregation2(Module):

    def __init__(self, in_features, out_features, b_dim, dropout):
        super(GraphAggregation2, self).__init__()
        self.sigmoid_linear = nn.Sequential(nn.Linear(in_features+b_dim, out_features),
                                            nn.Sigmoid())
        self.tanh_linear = nn.Sequential(nn.Linear(in_features+b_dim, out_features),
                                         nn.Tanh())
        self.dropout = nn.Dropout(dropout)

    def forward(self, inputs, activation):
        i = self.sigmoid_linear(inputs)
        j = self.tanh_linear(inputs)
        output = torch.sum(torch.mul(i, j), 1)
        output = activation(output) if activation is not None else output
        output = self.dropout(output)

        return output


class MultiDenseLayer(Module):
    def __init__(self, aux_unit, linear_units, activation=None, dropout_rate=0.):
        super(MultiDenseLayer, self).__init__()
        layers = []
        for c0, c1 in zip([aux_unit] + linear_units[:-1], linear_units):
            layers.append(nn.Linear(c0, c1))
            layers.append(nn.Dropout(dropout_rate))
            if activation is not None:
                layers.append(activation)
        self.linear_layer = nn.Sequential(*layers)

    def forward(self, inputs):
        h = self.linear_layer(inputs)
        return h



from jaqpotpy.datasets.dataset_base import ImageDataset
from typing import Iterable, Any, Optional, Callable
from PIL import Image
import os
import matplotlib.pyplot as plt
import inspect
import numpy as np
import pickle
from torch.utils.data import Dataset


IMG_EXTENSIONS = (".jpg", ".jpeg", ".png", ".ppm", ".bmp", ".pgm", ".tif", ".tiff", ".webp")


def pil_loader(path: str) -> Image.Image:
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, "rb") as f:
        img = Image.open(f)
        return img.convert("RGB")


def accimage_loader(path: str) -> Any:
    import accimage

    try:
        return accimage.Image(path)
    except OSError:
        # Potentially a decoding problem, fall back to PIL.Image
        return pil_loader(path)


def default_loader(path: str) -> Any:
    from torchvision import get_image_backend

    if get_image_backend() == "accimage":
        return accimage_loader(path)
    else:
        return pil_loader(path)


class TorchImageDataset(ImageDataset):
    """
    Reads images from a path and creates tensors ready for modelling. This is compatible with all transform
    methods provided by torchvision, as well as custom callable transformers.
    """

    def __init__(self, path=None, X: Optional[Iterable[Any]] = None, y: Optional[Iterable[Any]] = None,
                 featurizer: Callable = None, task: str = "regression", y_name: str = None,
                 images_name: str = None) -> None:
        super(ImageDataset, self).__init__(path=path, x_cols=images_name, y_cols=y_name)
        self._y = y_name
        self._x = X
        self._X = 'ImagePath'
        self.ys = y
        self._task = task
        self.featurizer: Callable = featurizer
        self.images = None
        self.data = None
        self.raw_img = None
        # self.create()

    def create(self):
        if not self.path:
            self.path = '/'
        if not self._x:
            # All files in the self.path with extentions defined in IMG_EXTENSIONS
            # will be selected as X
            self.images = [os.path.join(self.path, item) for item in os.listdir(self.path) if
                           item.endswith(IMG_EXTENSIONS)]
        else:
            self.images = [os.path.join(self.path, item) for item in self._x]

        self.raw_img = [default_loader(image) for image in self.images]

        if self.featurizer:
            self.data = tuple([(self.featurizer(image), y_i) for image, y_i in zip(self.raw_img, self.ys)])
        else:
            self.data = tuple([(image, y_i) for image, y_i in zip(self.raw_img, self.ys)])

        return self

    def show_image(self, idx, cmap=None):
        plt.imshow(self.raw_img[idx], cmap=cmap)
        plt.title('%i' % self.ys[idx])
        plt.show()

    def __get_X__(self):
        return self.images

    def __get_Y__(self):
        return self.ys

    def __get__(self):
        return self.df

    def __getitem__(self, idx):
        # print(self.df[self.X].iloc[idx].values)
        # print(type(self.df[self.X].iloc[idx].values))
        return self.data[idx]

    def __len__(self):
        return len(self.images)

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__


import numpy as np

from jaqpotpy.utils.types import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
from rdkit.Chem.AllChem import GetMACCSKeysFingerprint
from rdkit.Chem import DataStructs


class MACCSKeysFingerprint(MolecularFeaturizer):
  """MACCS Keys Fingerprint.
  The MACCS (Molecular ACCess System) keys are one of the most commonly used structural keys.
  Please confirm the details in [1]_, [2]_.
  Examples
  --------
  >>> import jaqpotpy as jp
  >>> smiles = 'CC(=O)OC1=CC=CC=C1C(=O)O'
  >>> featurizer = jp.descriptors.molecular.MACCSKeysFingerprint()
  >>> features = featurizer.featurize([smiles])
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (167,)
  References
  ----------
  .. [1] Durant, Joseph L., et al. "Reoptimization of MDL keys for use in drug discovery."
     Journal of chemical information and computer sciences 42.6 (2002): 1273-1280.
  .. [2] https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/MACCSkeys.py
  Note
  ----
  This class requires RDKit to be installed.
  """
  @property
  def __name__(self):
    return 'MACCSKeysFingerprint'

  def __getitem__(self):
    return self

  def __init__(self):
    """Initialize this featurizer."""
    self.calculator = None

  def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
    """
    Calculate MACCS keys fingerprint.
    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit Mol object
    Returns
    -------
    np.ndarray
      1D array of RDKit descriptors for `mol`. The length is 167.
    """
    if 'mol' in kwargs:
      datapoint = kwargs.get("mol")
      raise DeprecationWarning(
          'Mol is being phased out as a parameter, please pass "datapoint" instead.'
      )

    if self.calculator is None:
      try:
        self.calculator = GetMACCSKeysFingerprint
      except ModuleNotFoundError:
        raise ImportError("This class requires RDKit to be installed.")

    fp = self.calculator(datapoint)
    array = np.zeros((0,), dtype=np.int8)
    DataStructs.ConvertToNumpyArray(fp, array)
    return np.asarray(array)

  def _get_column_names(self, **kwargs) -> list:
    descriptors = []
    for i in range(167):
      descriptors.append("f" + str(i))
    return descriptors

  def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
    return self._featurize(datapoint, **kwargs)



from jaqpotpy.doa.doa import DOA
from jaqpotpy.descriptors.molecular import RDKitDescriptors, MACCSKeysFingerprint
from jaqpotpy.descriptors.base_classes import Featurizer #MolecularFeaturizer, MaterialFeaturizer,
from typing import Any, Iterable, Union, Dict
from jaqpotpy.datasets.image_datasets import default_loader
try:
    from pymatgen.core.structure import Lattice, Structure
except ModuleNotFoundError:
    Lattice = Structure = None
    pass
import pandas as pd

import numpy as np
import base64
import torch.nn.functional as nnf
import pickle
import os
import torch


class Model(object):
    _model: Any
    _doa: DOA
    _descriptors: Featurizer
    _preprocessors: []
    _preprocessor_names: []
    _preprocessor_y_names: []
    _preprocessors_y: []
    _X: Iterable[str]
    _Y: Iterable[str]
    _X_indices: Iterable[int]
    _prediction: Any
    _probability = None
    _external = None
    _smiles = None
    _external_feats = None
    _model_title = None
    _modeling_task = None
    _library = Iterable[str]
    _version = Iterable[str]
    _jaqpotpy_version = None
    _jaqpotpy_docker = None
    _optimizer = None

    @property
    def smiles(self):
        return self._smiles

    @smiles.setter
    def smiles(self, values):
        self._smiles = values

    @property
    def external(self):
        return self._external

    @property
    def descriptors(self):
        return self._descriptors

    @descriptors.setter
    def descriptors(self, value):
        self._descriptors = value

    @property
    def doa(self):
        return self._doa

    @doa.setter
    def doa(self, value):
        self._doa = value

    @property
    def preprocessing(self):
        return self._preprocessors

    @preprocessing.setter
    def preprocessing(self, value):
        self._preprocessors = value

    @property
    def preprocessing_y(self):
        return self._preprocessors_y

    @preprocessing_y.setter
    def preprocessing_y(self, value):
        self._preprocessors_y = value

    @property
    def preprocessor_names(self):
        return self._preprocessor_names

    @preprocessor_names.setter
    def preprocessor_names(self, value):
        self._preprocessor_names = value

    @property
    def preprocessor_y_names(self):
        return self._preprocessor_names

    @preprocessor_y_names.setter
    def preprocessor_y_names(self, value):
        self._preprocessor_y_names = value

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        self._X = value

    @property
    def Y(self):
        return self._Y

    @Y.setter
    def Y(self, value):
        self._Y = value

    @property
    def external(self) -> dict:
        return self._external

    @external.setter
    def external(self, value):
        self._external = value

    @property
    def external_feats(self) -> Iterable[str]:
        return self._external_feats

    @external_feats.setter
    def external_feats(self, value):
        self._external_feats = value

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def model_title(self):
        return self._model_title

    @model_title.setter
    def model_title(self, value):
        self._model_title = value

    @property
    def prediction(self):
        return self._prediction

    @prediction.setter
    def prediction(self, value):
        self._prediction = value

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, value):
        self._probability = value

    @property
    def library(self):
        return self._library

    @library.setter
    def library(self, value):
        self._library = value

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, value):
        self._optimizer = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def jaqpotpy_version(self):
        return self._jaqpotpy_version

    @property
    def modeling_task(self):
        return self._modeling_task

    @modeling_task.setter
    def modeling_task(self, value):
        self._modeling_task = value

    @jaqpotpy_version.setter
    def jaqpotpy_version(self, value):
        self._jaqpotpy_version = value

    @property
    def jaqpotpy_docker(self):
        return self._jaqpotpy_docker

    @jaqpotpy_docker.setter
    def jaqpotpy_docker(self, value):
        self._jaqpotpy_docker = value

    def __train__(self):
        raise NotImplemented("Not implemented")

    def __eval__(self):
        raise NotImplemented("Not implemented")

    def infer(self):
        raise NotImplemented("Not implemented")


class MolecularModel(Model):

    def __call__(self, smiles, external=None):
        self._smiles = smiles
        self._external = external
        self.infer()

    def save(self):
        if self.model_title:
            with open(self.model_title + ".jmodel", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_model" + ".jmodel", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def deploy_on_jaqpot(self, jaqpot, description, model_title: str = None):
        jaqpot.deploy_jaqpotpy_molecular_model(self, description=description, title=model_title)

    @classmethod
    def load_from_jaqpot(cls, jaqpot, id: str):
        try:
            model = jaqpot.get_raw_model_by_id(id)
            raw_model = base64.b64decode(model['actualModel'][0])
            model: MolecularModel = pickle.loads(raw_model)
            return model
        except Exception as e:

            pass

    def infer(self):
        self._prediction = []
        self._probability = []
        if self._smiles:
            if self._descriptors == "RDKitDescriptors":
                self._descriptors = RDKitDescriptors()
                data = self._descriptors.featurize_dataframe(self._smiles)
            elif self._descriptors == "MACCSKeysFingerprint":
                self._descriptors = MACCSKeysFingerprint()
                data = self._descriptors.featurize_dataframe(self._smiles)
            elif type(self._descriptors).__name__ == "Compose":
                self._descriptors.__name__ = "Compose"
                data = [self._descriptors(default_loader(image)) for image in self._smiles]
            else:
                data = self._descriptors.featurize_dataframe(self._smiles)

            if self.external:
                ext = pd.DataFrame.from_dict(self.external)
                data = pd.concat([data, ext], axis=1)
            graph_data_list = []

            # if self._descriptors.__name__ == 'MolGraphConvFeaturizer':
                # graph_data = data['MoleculeGraph']
            if self.library == ['torch_geometric', 'torch']:
                column = data.columns[0]
                for g in data[column].to_list():
                    from torch_geometric.data import Data
                    dat = Data(x=torch.FloatTensor(g.node_features)
                               , edge_index=torch.LongTensor(g.edge_index)
                               , edge_attr=g.edge_features
                               , num_nodes=g.num_nodes)
                    graph_data_list.append(dat)
                self._prediction = []

            if self._X == 'ImagePath':
                pass
            elif self._X != ['TorchMolGraph'] and self.X != ['OneHotSequence'] and self.X != ["SmilesImage"]:
                data = data[self._X].to_numpy()
            elif self._X == ['SmilesImage']:
                datas = []
                for v in data.values:
                    datas.append(v[0].transpose(2, 0, 1))
                data = np.array(datas)
            elif self.X == ['OneHotSequence']:
                data_list = []
                for d in data.values:
                    data_list.append(d[0])
                data = np.array(data_list)
            if self.doa:
                if self.doa.__name__ == 'SmilesLeverage':
                    self.doa.predict(self._smiles)
                else:
                    self.doa.predict(data)
            try:
                if self.preprocessing:
                    for f in self.preprocessing:
                        data = f.transform(data)
            except AttributeError as e:
                pass
            if self.library == ['sklearn']:
                preds = self.model.predict(data)
                for p in preds:
                    try:
                        if self.preprocessing_y:
                            for f in self.preprocessing_y:
                                p = f.inverse_transform(p.reshape(1, -1))
                    except AttributeError as e:
                        pass
                    self._prediction.append(p.tolist())
                try:
                    probs = self.model.predict_proba(data)
                    self._probability = [prob.tolist() for prob in probs]

                except AttributeError as e:
                    pass
            if self.library == ['torch_geometric', 'torch']:
                try:
                    self.model.eval()
                except AttributeError as e:
                    with open("./temp_model.pt", "wb") as f:
                        f.write(self.model)
                        f.close()
                    self.model = torch.jit.load("./temp_model.pt")
                    os.remove("./temp_model.pt")
                    self.model.eval()
                from torch_geometric.loader import DataLoader
                data_loader = DataLoader(graph_data_list, batch_size=1)
                for data in data_loader:
                    x = data.x
                    edge_index = data.edge_index
                    try:
                        edge_attributes = torch.Tensor(data.edge_attr[0])
                    except TypeError:
                        edge_attributes = None

                    if edge_attributes is not None:
                        pred = self.model(x, edge_index, edge_attributes, data.batch)
                    else:
                        pred = self.model(x, edge_index, data.batch)
                    if self.modeling_task == "classification":
                        for p in pred:
                            prob = nnf.softmax(p, dim=0)
                            self._probability.append(prob.detach().numpy().tolist())
                            # self._probability.append(p.detach().numpy().tolist())
                        pred = pred.argmax(dim=1)
                        preds = pred.detach().numpy()
                        for p in preds:
                            try:
                                if self.preprocessing_y:
                                    for f in self.preprocessing_y:
                                        p = f.inverse_transform(p)
                            except AttributeError as e:
                                pass
                            self._prediction.append(p.tolist())
                    else:
                        preds = pred.detach().numpy()
                        for p in preds:
                            self._prediction.append(p.tolist())
            if self.library == ['torch']:
                try:
                    self.model.eval()
                except AttributeError as e:
                    with open("./temp_model.pt", "wb") as f:
                        f.write(self.model)
                        f.close()
                    self.model = torch.jit.load("./temp_model.pt")
                    os.remove("./temp_model.pt")
                    self.model.eval()
                from torch.utils.data import DataLoader
                data_loader = DataLoader(data, batch_size=1)
                for data in data_loader:

                    pred = self.model(data.float())
                    if self.modeling_task == "classification":
                        for p in pred:
                            prob = nnf.softmax(p, dim=0)
                            self._probability.append(prob.detach().numpy().tolist())
                            # self._probability.append(p.detach().numpy().tolist())
                        pred = pred.argmax(dim=1)
                        preds = pred.detach().numpy()
                        for p in preds:
                            try:
                                if self.preprocessing_y:
                                    for f in self.preprocessing_y:
                                        p = f.inverse_transform(p)
                            except AttributeError as e:
                                pass
                            self._prediction.append([p.tolist()])
                    else:
                        preds = pred.detach().numpy()
                        for p in preds:
                            try:
                                if self.preprocessing_y:
                                    for f in self.preprocessing_y:
                                        p = f.inverse_transform(p)
                            except AttributeError as e:
                                pass
                            self._prediction.append([p.tolist()])
            # else:
            #     preds = self.model.predict(data)
            # for p in preds:
            #     self._prediction.append([p[0]])
            if type(self._descriptors).__name__ == "RDKitDescriptors":
                self._descriptors = "RDKitDescriptors"
            if type(self._descriptors).__name__ == "MACCSKeysFingerprint":
                self._descriptors = "MACCSKeysFingerprint"
            return self
        else:
            pass


class MaterialModel(Model):

    def __call__(self, compositions: Iterable[str] = None, structures: Union[Iterable[Structure], Iterable[Dict]] = None, external=None):
        if compositions:
            if structures:
                raise ValueError('Both compositions and structures were passed. Please provide one of them.')
            else:
                self.__type_mat__ = 'composition'
                self._materials = compositions
        else:
            self.__type_mat__ = 'structure'
            self._materials = structures
        self._external = external
        self.infer()

    def save(self):
        if self._model_title:
            with open(self._model_title + ".jmodel", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_model" + ".jmodel", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def infer(self):
        self._prediction = []
        self._probability = []
        if self._materials:
            ext = pd.DataFrame.from_dict(self.external)
            data = self._descriptors.featurize_dataframe(self._materials)
            data = pd.concat([data, ext], axis=1)
            graph_data_list = []
            if self._descriptors.__name__ == 'CrystalGraphCNN':
                for g in data['MaterialGraph'].to_list():
                    from torch_geometric.data import Data
                    dat = Data(x=torch.FloatTensor(g.node_features)
                               , edge_index=torch.LongTensor(g.edge_index)
                               , edge_attr=g.edge_features
                               , num_nodes=g.num_nodes)
                    graph_data_list.append(dat)
                self._prediction = []

            # if self.doa:
            #     self.doa_m = self.doa.fit(X=self.dataset.__get_X__())

            try:
                if self.preprocessing:
                    for f in self.preprocessing:
                        data = f.transform(data)
            except AttributeError as e:
                pass
            if self.library == ['sklearn']:
                preds = self.model.predict(data)
                for p in preds:
                    try:
                        if self.preprocessing_y:
                            for f in self.preprocessing_y:
                                p = f.inverse_transform(data)
                    except AttributeError as e:
                        pass
                    self._prediction.append(p)
                try:
                    probs = self.model.predict_proba(data)
                    for prob in probs:
                        self._probability.append(prob)
                except AttributeError as e:
                    pass
            if self.library == ['torch_geometric', 'torch']:
                self.model.eval()
                # self.model.no_grad()
                # torch.no_grad()
                from torch_geometric.loader import DataLoader
                data_loader = DataLoader(graph_data_list, batch_size=1)
                for g in data_loader:
                    pred = self.model(g)
                    if self.modeling_task == "classification":
                        for p in pred:
                            self._probability.append(p.detach().numpy().tolist())
                        pred = pred.argmax(dim=1)
                        preds = pred.detach().numpy()
                        for p in preds:
                            try:
                                if self.preprocessing_y:
                                    for f in self.preprocessing_y:
                                        p = f.inverse_transform(data)
                            except AttributeError as e:
                                pass
                            self._prediction.append([p.tolist()])
                    else:
                        preds = pred.detach().numpy()
                        for p in preds:
                            self._prediction.append(p.tolist())
            if self.library == ['torch']:
                self.model.eval()
                # self.model.no_grad()
                # torch.no_grad()
                from torch.utils.data import DataLoader
                data_loader = DataLoader(data, batch_size=1)
                for g in data_loader:
                    pred = self.model(g.float())
                    if self.modeling_task == "classification":
                        for p in pred:
                            self._probability.append(p.detach().numpy().tolist())
                        pred = pred.argmax(dim=1)
                        preds = pred.detach().numpy()
                        for p in preds:
                            try:
                                if self.preprocessing_y:
                                    for f in self.preprocessing_y:
                                        p = f.inverse_transform(data)
                            except AttributeError as e:
                                pass
                            self._prediction.append([p.tolist()])
                    else:
                        preds = pred.detach().numpy()
                        for p in preds:
                            self._prediction.append(p.tolist())
            # else:
            #     preds = self.model.predict(data)
            # for p in preds:
            #     self._prediction.append([p[0]])
            return self
        else:
            pass



import unittest
from jaqpotpy.jaqpot import Jaqpot

from jaqpotpy.models import MolecularModel
from jaqpotpy.models.evaluator import GenerativeEvaluator
from jaqpotpy.models.generative.molecular_metrics import diversity_scores \
    , drugcandidate_scores, synthetic_accessibility_score_scores, valid_mean \
    , quantitative_estimation_druglikeness_scores, novel_score \
    , water_octanol_partition_coefficient_scores, unique_total_score, valids, valid_scores


class TestEvaluators(unittest.TestCase):

    def setUp(self) -> None:
        self.jaqpot = Jaqpot()
        self.jaqpot.set_api_key(
            "eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJ3Ujh3X1lGOWpKWFRWQ2x2VHF1RkswZkctQXROQUJsb3FBd0N4MmlTTWQ4In0.eyJleHAiOjE2NjY1MjU0MzQsImlhdCI6MTY2NjM1MjYzNCwiYXV0aF90aW1lIjoxNjY2MzUyNjM0LCJqdGkiOiIxODk2ZWQ3ZC1lMmM1LTQ2ZWItOTE3MS1mYzQyYWRiYzVkMmUiLCJpc3MiOiJodHRwczovL2xvZ2luLmphcXBvdC5vcmcvYXV0aC9yZWFsbXMvamFxcG90IiwiYXVkIjoiYWNjb3VudCIsInN1YiI6IjI0MjVkNzYwLTAxOGQtNDA4YS1hZTBiLWNkZTRjNTYzNTRiOSIsInR5cCI6IkJlYXJlciIsImF6cCI6ImphcXBvdC11aS1jb2RlIiwibm9uY2UiOiI0NDAxMTk1NDUyNGQzMTk5ZGIxNDZkYjAwMWViZGZiYzNlbFFSS2JwWiIsInNlc3Npb25fc3RhdGUiOiIzMGVlZTc1YS03NDg2LTRjN2MtYjI1Ni1kNmM0ZjIxOGI2MDQiLCJhY3IiOiIxIiwiYWxsb3dlZC1vcmlnaW5zIjpbIicqJyIsIioiXSwicmVhbG1fYWNjZXNzIjp7InJvbGVzIjpbIm9mZmxpbmVfYWNjZXNzIiwidW1hX2F1dGhvcml6YXRpb24iXX0sInJlc291cmNlX2FjY2VzcyI6eyJhY2NvdW50Ijp7InJvbGVzIjpbIm1hbmFnZS1hY2NvdW50IiwibWFuYWdlLWFjY291bnQtbGlua3MiLCJ2aWV3LXByb2ZpbGUiXX19LCJzY29wZSI6Im9wZW5pZCBqYXFwb3QtYWNjb3VudHMgZW1haWwgcHJvZmlsZSB3cml0ZSByZWFkIiwic2lkIjoiMzBlZWU3NWEtNzQ4Ni00YzdjLWIyNTYtZDZjNGYyMThiNjA0IiwiZW1haWxfdmVyaWZpZWQiOmZhbHNlLCJuYW1lIjoiUGFudGVsaXMgS2FyYXR6YXMiLCJwcmVmZXJyZWRfdXNlcm5hbWUiOiJwYW50ZWxpc3BhbmthIiwiZ2l2ZW5fbmFtZSI6IlBhbnRlbGlzIiwiZmFtaWx5X25hbWUiOiJLYXJhdHphcyIsImVtYWlsIjoicGFudGVsaXNwYW5rYUBnbWFpbC5jb20ifQ.GFFG1HIKcYsdVqAFSavL_1SgLN0GJWffL3yOahUV_4h2JmqYx5LkzELAk9EfLDroqzxclcl0uBEutrSKIPK2-PSN-8cl_4UY-9__E2F6iG7OR_wOtW5vOY54L5p8LqC0nnbrck_U59sVz9k-zzIsSgmV1NJwDxduL8nzN5FQBjnUW8ZAsMXx1AhK23-x3KPe2L1s2lAq4JE7azXL7ahu7tdSAaKsJRzF0ZdSwUCWic_awc5nNxjspoToK-OrC8-0cDx1cDgX30-L1Wy2gTMBzjW4gHkQ-uhKZ9-yVa_bWK162x0JRGzE_RzZf9DiPebE-jCShxTcuK_8pw1YfoILyw")
        self.mols_mocking_genereted = [
            "CCOc1cccc(NC(=O)NCc2ccc(N3CCSCC3)cc2)c1",
            "Cc1cccc(NC(=O)CN2CCN(c3ccc4c(c3)OCCO4)C2=O)n1",
            "C=C(C)C(=O)N[C@H](C)c1nc2ccccc2n1CCC(=O)N1CCCCCC1",
            "CCOC[C@H]1CC[NH+](Cc2ccc(-c3nc4ccccc4s3)o2)C1",
            "CCOC(=O)[C@]1(Cc2cccc(Cl)c2)CCCN(C(=O)c2ccnn2C)C1",
            "Cc1ccc([N+](=O)[O-])cc1NC(=O)C(=O)N1CC[C@H]([NH+]2CCCC2)C1",
            "CCOCCCNC(=O)N[C@@H]1CCC[C@@H](CC)C1",
            "O=C(Cc1cccc(F)c1F)Nc1cccc(Br)n1",
            "COc1ccccc1NC(=O)[C@@H]1CCCN(C(=O)Nc2cccs2)C1",
            "C[C@H]1CCC[C@](C#N)([C@]2(O)CCCCC2(C)C)C1",
            "O=C(NCc1ccc([N+]2=CCCC2)cc1)NC1(c2ccc(Cl)cc2)CC1",
            "CCCC(=O)N[C@@H]1CCC[NH+](Cc2ncccc2C)C1",
            "O=C(NCc1cccs1)C1(c2cccc(Cl)c2)CCC1",
            "C[C@H]1CCC[C@H](NC(=O)[C@@H](C)Sc2ncn[nH]2)[C@@H]1C",
            "COc1ccc([C@@H]([NH2+]Cc2ccc(Cl)nc2)c2ccc(F)cc2)cc1",
            "COc1cc(NC(=O)[C@H](C)Sc2ccccc2Cl)cc(OC)c1",
            "CCN1CCC(=NNC(=O)c2ccccc2)CC1",
            "CCCOc1ccc(Br)cc1C[NH+]1CCC([C@@H](C)O)CC1",
            "Cc1cc2n(C[C@H](O)CO[C@H](c3ccccc3)c3ccccc3C)c(=O)c3ccccc3n2n1",
            "CCc1ncc(CN(C)C(=O)Nc2c(C)ccc([N+](=O)[O-])c2C)s1",
            "c1ccc2nc(NCCCc3nc4ccccc4[nH]3)cnc2c1",
            "Cc1c([C@H](C)[NH2+]Cc2cccn2C)cnn1C",
            "CC(=O)N[C@@H](C(=O)NC1COC1)C(C)C",
            "O=C(Nc1ccccc1F)c1cc2ccccc2c2cccnc12",
            "Cc1ccccc1N1C(=O)/C(=C/c2cccn2-c2cccc([N+](=O)[O-])c2)C([O-])=NC1=S",
            "COCCN1C[C@H](C(=O)N(Cc2cccc(Cl)c2)C(C)C)CC1=O",
            "COc1ccc(NC(=O)N2CCN(C(=O)Cc3csc4ccccc34)CC2)cc1OC",
            "C#CCN(C[C@H]1CCCO1)C(=O)N[C@@H](C)c1cccc([N+](=O)[O-])c1",
            "Cc1cccc(C2=CCN(C(=O)Nc3ccc(C(N)=O)c(Cl)c3)CC2)c1",
            "COCCCN1C(=O)c2ccc(C(=O)Nc3nc(-c4ccc(C)cc4)cs3)cc2C1=O",
            "O=S(=O)(Nc1ccc(N2CCCS2(=O)=O)cc1)c1ccc(F)c(Cl)c1",
            "O=C1/C(=C/c2ccccc2)Oc2c1ccc1c2CN(Cc2cccs2)CO1",
            "CC[C@@H](C)[C@@H](NC(=O)c1cccc(F)c1)C(=O)N=c1[nH]c2ccccc2[nH]1",
            "Cc1c(F)cc(N)cc1S(=O)(=O)N[C@@H](C)C1CC1",
            "Cc1ccc(Cn2ncc3c(N)ncnc32)cc1",
            "CCOC(=O)C(C)(C)c1nc(-c2ccccc2)no1",
            "CCOC(=O)c1sc(/C=C/c2nc3c(s2)CCC3)nc1C",
            "C[C@@H]1CC[C@@H]([NH2+]C2CCC(NS(C)(=O)=O)CC2)c2ccccc21",
            "CN(C)S(=O)(=O)c1ccc(C(=O)N(C(=O)N2CCCCC2)c2ccccc2)cc1",
            "CC[NH2+][C@@H](CC)c1ccccc1OCc1cccc(F)c1",
            "C=CCOC(=O)C1=C(C)N=C2S[C@H](C)C(=O)N2[C@H]1c1ccc(F)cc1",
            "CCC[NH2+][C@]1(C(=O)OCC)CC[C@H](n2cc(Cl)c(C)n2)C1",
            "Cc1cc(NC(=O)CSc2nnc3c4ccccc4n(C)c3n2)ccc1Br",
            "CCOC(=O)c1sc(NC(=O)c2ccc(-n3c(C)nc4ccccc4c3=O)cc2)cc1C",
            "CC#CCC(=O)C1([NH+](CC)CC)CCCC1",
            "O=C(NCCCc1ccccc1)C1CCN(C(=O)[C@@H]2CC(=O)N(c3ccccc3)C2)CC1",
            "O=Cc1ccc(OCc2ccn(-c3cccc(F)c3)n2)cc1",
            "Clc1ccccc1Cn1ccnc1",
                     ]


        self.mols_no_valids = [
            "CCOc1ccccAdsfvSdSCC3)cc2)c1",
            "Cc1cccc(NC(=O)CN2CCN(c3ccc4c(c3)OCCO4)C2=O)n1",
            "C=C(C)C(=O)N[C@H](C)c1nc2ccccc2n1CCC(=O)N1CCCCCC1",
            "CCOC[C@H]1CC[NH+](Cc2ccc(-c3nc4ccccc4s3)o2)C1",
            "CCOC(=O)[C@]1(Cc2cccc(Cl)c2)CCCN(C(=O)c2ccnn2C)C1",
            "Cc1ccc([N+](=O)[O-])cc1NC(=O)C(=O)N1CC[C@H]([NH+]2CCCC2)C1",
            "CCOCCCNC(=O)N[C@@H]1CCC[C@@H](CC)C1",
            "O=C(Cc1cccc(F)c1F)Nc1cccc(Br)n1",
            "COc1ccccc1NC(=O)[C@@H]1CCCN(C(=O)Nc2cccs2)C1",
            "C[C@H]1CCC[C@](C#N)([C@]2(O)CCCCC2(C)C)C1",
            "O=C(NCc1ccc([N+]2=CCCC2)cc1)NC1(c2ccc(Cl)cc2)CC1",
            "CCCC(=O)N[C@@H]1CCC[NH+](Cc2ncccc2C)C1",
            "O=C(NCc1cccs1)C1(c2cccc(Cl)c2)CCC1",
            "C[C@H]1CCC[C@H](NC(=O)[C@@H](C)Sc2ncn[nH]2)[C@@H]1C",
            "COc1ccc([C@@H]([NH2+]Cc2ccc(Cl)nc2)c2ccc(F)cc2)cc1",
            "COc1cc(NC(=O)[C@H](C)Sc2ccccc2Cl)cc(OC)c1",
            "CCN1CCCadsfijsldc2)CC1",
            "CCCOc1ccc(Br)cc1C[NH+]1CCC([C@@H](C)O)CC1",
            "Cc1cc2n(C[C@H](O)CO[C@H](c3ccccc3)c3ccccc3C)c(=O)c3ccccc3n2n1",
            "CCc1ncc(CN(C)C(=O)Nc2c(C)ccc([N+](=O)[O-])c2C)s1",
            "c1ccc2nc(NCCCc3nc4ccccc4[nH]3)cnc2c1",
            "Cc1c([C@H](C)[NH2+]Cc2cccn2C)cnn1C",
            "CC(=O)N[C@@H](C(C)C",
            "",
            "Cc1ccccc1N1C(=O)/C(=C/c2cccn2-c2cccc([N+](=O)[O-])c2)C([O-])=NC1=S",
            "COCCN1C[C@H](C(=O)N(Cc2cccc(Cl)c2)C(C)C)CC1=O",
            "COc1ccc(NC(=O)N2CCN(C(=O)Cc3csc4ccccc34)CC2)cc1OC",
            "C#CCN(C[C@H]1CCCO1)C(=O)N[C@@H](C)c1cccc([N+](=O)[O-])c1",
            "Cc1cccc(C2=CCN(C(=O)Nc3ccc(C(N)=O)c(Cl)c3)CC2)c1",
            "COCCCN1C(=O)c2ccc(C(=O)Nc3nc(-c4ccc(C)cc4)cs3)cc2C1=O",
            "O=S(=O)(Nc1ccc(N2CCCS2(=O)=O)cc1)c1ccc(F)c(Cl)c1",
            "O=C1/C(=C/c2ccccc2)Oc2c1ccc1c2CN(Cc2cccs2)CO1",
            "CC[C@@H](C)[C@@H](NC(=O)c1cccc(F)c1)C(=O)N=c1[nH]c2ccccc2[nH]1",
            "Cc1c(F)cc(N)cc1S(=O)(=O)N[C@@H](C)C1CC1",
            "Cc1ccc(Cn2ncc3c(N)ncnc32)cc1",
            "CCOC(=O)C(C)(C)c1nc(-c2ccccc2)no1",
            "CCOC(=O)c1sc(/C=C/c2nc3c(s2)CCC3)nc1C",
            "C[C@@H]1CC[C@@H]([NH2+]C2CCC(NS(C)(=O)=O)CC2)c2ccccc21",
            "CN(C)S(=O)(=O)c1ccc(C(=O)N(C(=O)N2CCCCC2)c2ccccc2)cc1",
            "CC[NH2+][C@@H](CC)c1ccccc1OCc1cccc(F)c1",
            "C=CCOC(=O)C1=C(C)N=C2S[C@H](C)C(=O)N2[C@H]1c1ccc(F)cc1",
            "CCC[NH2+][C@]1(C(=O)OCC)CC[C@H](n2cc(Cl)c(C)n2)C1",
            "Cc1cc(NC(=O)CSc2nnc3c4ccccc4n(C)c3n2)ccc1Br",
            "CCOC(=O)c1sc(NC(=O)c2ccc(-n3c(C)nc4ccccc4c3=O)cc2)cc1C",
            "CC#CCC(=O)C1([NH+](CC)CC)CCCC1",
            "O=C(NCCCc1ccccc1)C1CCN(C(=O)[C@@H]2CC(=O)N(c3ccccc3)C2)CC1",
            "O=Cc1ccc(OCc2ccn(-c3cccc(F)c3)n2)cc1",
            "Clc1ccccc1Cn1ccnc1",
                     ]


        self.mols_test = [
            "CC(C)(C)c1ccc2occ(CC(=O)Nc3ccccc3F)c2c1",
            "C[C@@H]1CC(Nc2cncc(-c3nncn3C)c2)C[C@@H](C)C1",
            "N#Cc1ccc(-c2ccc(O[C@@H](C(=O)N3CCCC3)c3ccccc3)cc2)cc1",
            "CCOC(=O)[C@@H]1CCCN(C(=O)c2nc(-c3ccc(C)cc3)n3c2CCCCC3)C1",
            "N#CC1=C(SCC(=O)Nc2cccc(Cl)c2)N=C([O-])[C@H](C#N)C12CCCCC2",
            "CC[NH+](CC)[C@](C)(CC)[C@H](O)c1cscc1Br",
            "COc1ccc(C(=O)N(C)[C@@H](C)C/C(N)=N/O)cc1O",
            "O=C(Nc1nc[nH]n1)c1cccnc1Nc1cccc(F)c1",
            "Cc1c(/C=N/c2cc(Br)ccn2)c(O)n2c(nc3ccccc32)c1C#N",
            "C[C@@H]1CN(C(=O)c2cc(Br)cn2C)CC[C@H]1[NH3+]",
            "CCOc1ccc(OCC)c([C@H]2C(C#N)=C(N)N(c3ccccc3C(F)(F)F)C3=C2C(=O)CCC3)c1",
            "Cc1ccc2nc(S[C@H](C)C(=O)NC3CCC(C)CC3)n(C)c(=O)c2c1",
            "O=C(N1CCc2c(F)ccc(F)c2C1)C1(O)Cc2ccccc2C1",
            "Cc1ccccc1C(=O)N1CCC2(CC1)C[C@H](c1ccccc1)C(=O)N2C",
            "CCCc1cc(NC(=O)CN2C(=O)NC3(CCC(C)CC3)C2=O)n(C)n1",
            "CC(C)Cc1nc(SCC(=O)NC[C@@H]2CCCO2)c2c(=O)n(C)c(=O)n(C)c2n1",
            "Cc1ccc(CNC(=O)c2ccccc2NC(=O)[C@@H]2CC(=O)N(c3ccc(C)cc3)C2)cc1",
            "CCCCC(=O)NC(=S)Nc1ccccc1C(=O)N1CCOCC1",
            "Cc1c(NC(=O)CSc2nc3sc4c(c3c(=O)[nH]2)CCCC4)c(=O)n(-c2ccccc2)n1C",
            "CC(C)[C@@H](Oc1cccc(Cl)c1)C(=O)N1CCC(n2cccn2)CC1",
            "CCN(CC)C(=O)C[C@@H](C)[NH2+][C@H](C)c1cccc(F)c1F",
            "Cc1nc2c(c(Nc3ncc(C)s3)n1)CCN(C(=O)CCc1ccccc1)C2",
            "O=C(NCCNC(=O)N1C[C@H]2CC=CC[C@@H]2C1)c1cccnc1",
            "O=c1n(CCO)c2ccccc2n1CCO",
            "COC(=O)Cc1csc(NC(=O)Cc2coc3cc(C)ccc23)n1",
            "Cc1ccc(N2CC[C@@H](NS(=O)(=O)c3ccccc3C)C2=O)cc1C",
            "CC[C@H](C)C[C@@H](C)NC(=O)N1CCN(CC(=O)NC2CC2)CC1",
            "CC(=O)Nc1c2n(c3ccccc13)C[C@](C)(C(=O)NC1CCCCC1)N(C1CCCCC1)C2=O",
            "N#Cc1ccncc1NC[C@@H]1C[C@@]12CCc1ccccc12",
            "Cc1cccn2c(=O)c(C(=O)NC[C@H]3CCO[C@@H]3C(C)C)cnc12",
            "CNC(=O)c1ccc(/C=C/C(=O)Nc2c(C)cc(C)nc2Cl)cc1",
            "CC1=C(CNC(=O)c2cc(-c3ccccc3)nc3c2CNN3C(C)C)CN=N1",
            "C[C@@H](NC(=O)COC(=O)/C=C/c1ccc(Cl)cc1)c1ccccc1",
            "CCc1ccc(N(Cc2ccc(C)s2)C(=O)c2ccc(=O)n(C)n2)cc1",
            "CCOC(=O)c1nnc2ccccc2c1N1CC[C@@H]([NH+](CC)CC)C1",
            "Cc1ccc(C#N)cc1S(=O)(=O)NCc1ccnc(OC(C)(C)C)c1",
            "O=C(O[C@H]1CCOC1)C1(c2ccc(Cl)c(Cl)c2)CCC1",
            "CCC[NH2+][C@@H]1COC[C@H]1C(=O)NCc1cscc1C",
            "O=C(NCc1nccc2ccccc12)c1ccc[nH]c1=O",
            "CC(=O)c1ccc(S(=O)(=O)N2CCCC[C@H]2C)cc1",
        ]

    # def test_load_jaqpot_model(self):
    #     model = MolecularModel.load_from_jaqpot(jaqpot=self.jaqpot, id="iaFePw8jwhHSbzg42acr")
    #     print(model.library)
    #     print(model.version)
    #     import torch
    #     print(torch.__version__)
    #     import torch_geometric
    #     print(torch_geometric.__version__)
    #     model(self.mols)
    #     print(model.prediction)

    def test_generative_evaluator(self):
        from rdkit import Chem
        gen_eval = GenerativeEvaluator()
        gen_eval.register_scoring_function("Valid all", valid_mean)
        gen_eval.register_scoring_function("QED", quantitative_estimation_druglikeness_scores)
        gen_eval.register_scoring_function("Synthetic Accessibility", synthetic_accessibility_score_scores)
        gen_eval.register_scoring_function("Novel", novel_score)
        gen_eval.register_scoring_function("Unique", unique_total_score)
        # gen_eval.register_scoring_function("Diversity", diversity_scores)
        gen_eval.register_scoring_function("Water Oct", water_octanol_partition_coefficient_scores)
        gen_eval.register_scoring_function("Drugcandidate Scores", drugcandidate_scores)

        # gen_eval.register_scoring_function("Novel all", novel_score)
        gen_eval.register_dataset(self.mols_test)
        mols = []
        for i in self.mols_mocking_genereted:
            mols.append(Chem.MolFromSmiles(i))
        rew = gen_eval.get_reward(mols)
        print(rew)

    def test_generative_evaluator_no_valids(self):
        from rdkit import Chem
        gen_eval = GenerativeEvaluator()
        gen_eval.register_scoring_function("Valid all", valid_mean)
        gen_eval.register_scoring_function("QED", quantitative_estimation_druglikeness_scores)
        gen_eval.register_scoring_function("Synthetic Accessibility", synthetic_accessibility_score_scores)
        gen_eval.register_scoring_function("Novel", novel_score)
        gen_eval.register_scoring_function("Unique", unique_total_score)
        # gen_eval.register_scoring_function("Diversity", diversity_scores)
        gen_eval.register_scoring_function("Water Oct", water_octanol_partition_coefficient_scores)
        gen_eval.register_scoring_function("Drugcandidate Scores", drugcandidate_scores)

        # gen_eval.register_scoring_function("Novel all", novel_score)
        gen_eval.register_dataset(self.mols_test)
        mols = []
        for i in self.mols_no_valids:
            try:
                mol = Chem.MolFromSmiles(i)
                mols.append(mol)
            except Exception as e:
                mols.append(None)
        rew = gen_eval.get_reward(mols)


    def test_generative_evaluator_scores(self):
        from rdkit import Chem
        gen_eval = GenerativeEvaluator()
        gen_eval.register_scoring_function("Valid all", valid_mean)
        gen_eval.register_scoring_function("QED", quantitative_estimation_druglikeness_scores)
        gen_eval.register_scoring_function("Synthetic Accessibility", synthetic_accessibility_score_scores)
        gen_eval.register_scoring_function("Novel", novel_score)
        gen_eval.register_scoring_function("Unique", unique_total_score)
        gen_eval.register_scoring_function("Diversity", diversity_scores)
        gen_eval.register_scoring_function("Water Oct", water_octanol_partition_coefficient_scores)
        gen_eval.register_scoring_function("Drugcandidate Scores", drugcandidate_scores)

        # gen_eval.register_scoring_function("Novel all", novel_score)
        gen_eval.register_dataset(self.mols_test)
        mols = []
        for i in self.mols_no_valids:
            try:
                mol = Chem.MolFromSmiles(i)
                mols.append(mol)
            except Exception as e:
                mols.append(None)
        rew = gen_eval.get_reward(mols)

    def test_generative_evaluator_scores_valids(self):
        from rdkit import Chem
        gen_eval = GenerativeEvaluator()
        gen_eval.register_scoring_function("Valids", valid_scores)
        gen_eval.register_scoring_function("QED", quantitative_estimation_druglikeness_scores)

        # gen_eval.register_scoring_function("Novel all", novel_score)
        gen_eval.register_dataset(self.mols_test)
        mols = []
        for i in self.mols_no_valids:
            try:
                mol = Chem.MolFromSmiles(i)
                mols.append(mol)
            except Exception as e:
                mols.append(None)
        rew = gen_eval.get_reward(mols)




# gen_eval = GenerativeEvaluator()
#
#
# def jaqpotpy_activity(mols):
#     from jaqpotpy.models import MolecularModel
#     from jaqpotpy.jaqpot import Jaqpot
#     jaqpot = Jaqpot()
#     jaqpot.login("pantelispanka", "kapan2")
#     model = MolecularModel().load_from_jaqpot("u60obQ1Y23e7GL2SQsYe")
#     model(mols)
#     pred = model.prediction
#     percentage = 0.99
#     return percentage
#
# gen_eval.register_scoring_function("Jaqpotpy_covid_activity", jaqpotpy_activity)
#
#
# def jaqpotpy_tox(mols):
#     from jaqpotpy.models import MolecularModel
#     tox_model = MolecularModel().load_from_jaqpot("Jaqpotid")
#     tox_model(mols)
#     pred = tox_model.prediction
#     percentage = 0.60
#     return percentage
#
# gen_eval.register_scoring_function("jaqpotpy_tox", jaqpotpy_tox)
# gen_eval.get_reward(mols="")
#
#
# def jaqpotpy_pipeline(mols):
#     from jaqpotpy.models import MolecularModel
#     model = MolecularModel().load_from_jaqpot("Jaqpotid")
#     tox_model = MolecularModel().load_from_jaqpot("Jaqpotid")
#
# from jaqpotpy.models.generative.molecular_metrics import novel_score
# gen_eval.register_scoring_function("novel", novel_score)



"""
Dataset base classes
"""
from typing import Any
import inspect
from typing import Iterable
import pickle


class BaseDataset(object):
    """
    Astract class for datasets
    """
    def __init__(self, path=None, x_cols=None, y_cols=None) -> None:
        self._Y = None
        self._X = None
        self._dataset_name = None
        self._df = None
        self._x_cols_all = None
        self.path = path
        self.x_cols = x_cols
        self.y_cols = y_cols
        self._task = "regression"
        self.featurizer = None
        self._featurizer_name = None
        self._external = None

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, value):
        self._task = value

    @property
    def dataset_name(self):
        return self._dataset_name

    @dataset_name.setter
    def dataset_name(self, value):
        self._dataset_name = value

    @property
    def featurizer_name(self) -> Iterable[Any]:
        return self.featurizer.__name__

    @property
    def x_colls_all(self) -> Iterable[str]:
        return self._x_cols_all

    @x_colls_all.setter
    def x_colls_all(self, value):
        self._x_cols_all = value

    @property
    def X(self) -> Iterable[str]:
        return self._x

    @X.setter
    def X(self, value):
        self._x = value

    @property
    def y(self) -> Iterable[str]:
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    @property
    def external(self) -> Iterable[str]:
        return self._external

    @external.setter
    def external(self, value):
        self._external = value

    @property
    def df(self) -> Any:
        return self._df

    @df.setter
    def df(self, value):
        self._df = value

    @featurizer_name.setter
    def featurizer_name(self, value):
        self._featurizer_name = value

    def create(self):
        raise NotImplementedError("Need implementation")

    def __repr__(self) -> str:
        # args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        # args_names = [arg for arg in args_spec.args if arg != 'self']
        # args_info = ''
        # for arg_name in args_names:
        #   value = self.__dict__[arg_name]
        #   # for str
        #   if isinstance(value, str):
        #     value = "'" + value + "'"
        #   # for list
        return self.__class__.__name__


class MolecularDataset(BaseDataset):
    def __init__(self, path=None, smiles_col=None, x_cols=None, y_cols=None, smiles=None) -> None:
        self.smiles = smiles
        self._smiles_strings = None
        self.smiles_col = smiles_col
        super().__init__(path, x_cols, y_cols)

    @property
    def smiles_strings(self) -> Iterable[str]:
        return self._smiles_strings

    @smiles_strings.setter
    def smiles_strings(self, value):
        self._smiles_strings = value

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdata", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_dataset" + ".jdata", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)


class MaterialDataset(BaseDataset):
    def __init__(self, path=None, materials_col=None, x_cols=None, y_cols=None, materials=None) -> None:
        self.materials = materials
        self._materials_strings = None
        self.materials_col = materials_col
        super().__init__(path, x_cols, y_cols)


    @property
    def materials_strings(self) -> Iterable[str]:
        return self._materials_strings

    @materials_strings.setter
    def materials_strings(self, value):
        self._materials_strings = value

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdata", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_dataset" + ".jdata", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

class ImageDataset(BaseDataset):
    def __init__(self, path=None,  x_cols=None, y_cols=None) -> None:
        super().__init__(path=path, x_cols=x_cols, y_cols=y_cols)

    def save(self):
        if self._dataset_name:
            with open(self._dataset_name + ".jdata", 'wb') as f:
                pickle.dump(self, f)
        else:
            with open("jaqpot_dataset" + ".jdata", 'wb') as f:
                pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)


"""
Topological fingerprints.
"""
from typing import Dict

import numpy as np

from jaqpotpy.utils.types import RDKitMol
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer


class TopologicalFingerprint(MolecularFeaturizer):
    """Circular (Morgan) fingerprints.
  Extended Connectivity Circular Fingerprints compute a bag-of-words style
  representation of a molecule by breaking it into local neighborhoods and
  hashing into a bit vector of the specified size. It is used specifically
  for structure-activity modelling. See [1]_ for more details.
  References
  ----------
  .. [1] Rogers, David, and Mathew Hahn. "Extended-connectivity fingerprints."
     Journal of chemical information and modeling 50.5 (2010): 742-754.
  Note
  ----
  This class requires RDKit to be installed.
  Examples
  --------
  >>> import jaqpotpy as jt
  >>> from rdkit import Chem
  >>> smiles = ['C1=CC=CC=C1']
  >>> # Example 1: (size = 2048, radius = 4)
  >>> featurizer = jt.descriptors.TopologicalFingerprint(size=2048, radius=4)
  >>> features = featurizer.featurize(smiles)
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (2048,)
  >>> # Example 2: (size = 2048, radius = 4, sparse = True, smiles = True)
  >>> featurizer = jt.descriptors.TopologicalFingerprint(size=2048, radius=8,
  ...                                          sparse=True, smiles=True)
  >>> features = featurizer.featurize(smiles)
  >>> type(features[0]) # dict containing fingerprints
  <class 'dict'>
  """

    @property
    def __name__(self):
        return 'TopologicalFingerprint'

    def __init__(self,
                 radius: int = 2,
                 size: int = 2048,
                 chiral: bool = False,
                 bonds: bool = True,
                 features: bool = False,
                 sparse: bool = False,
                 smiles: bool = False):
        """
    Parameters
    ----------
    radius: int, optional (default 2)
      Fingerprint radius.
    size: int, optional (default 2048)
      Length of generated bit vector.
    chiral: bool, optional (default False)
      Whether to consider chirality in fingerprint generation.
    bonds: bool, optional (default True)
      Whether to consider bond order in fingerprint generation.
    features: bool, optional (default False)
      Whether to use feature information instead of atom information; see
      RDKit docs for more info.
    sparse: bool, optional (default False)
      Whether to return a dict for each molecule containing the sparse
      fingerprint.
    smiles: bool, optional (default False)
      Whether to calculate SMILES strings for fragment IDs (only applicable
      when calculating sparse fingerprints).
    """
        self.radius = radius
        self.size = size
        self.chiral = chiral
        self.bonds = bonds
        self.features = features
        self.sparse = sparse
        self.smiles = smiles

    def __getitem__(self):
        return self

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate circular fingerprint.
    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit Mol object
    Returns
    -------
    np.ndarray
      A numpy array of circular fingerprint.
    """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdMolDescriptors
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.sparse:
            info: Dict = {}
            fp = rdMolDescriptors.GetMorganFingerprint(
                datapoint,
                self.radius,
                useChirality=self.chiral,
                useBondTypes=self.bonds,
                useFeatures=self.features,
                bitInfo=info)
            fp = fp.GetNonzeroElements()  # convert to a dict

            # generate SMILES for fragments
            if self.smiles:
                fp_smiles = {}
                for fragment_id, count in fp.items():
                    root, radius = info[fragment_id][0]
                    env = Chem.FindAtomEnvironmentOfRadiusN(datapoint, radius, root)
                    frag = Chem.PathToSubmol(datapoint, env)
                    smiles = Chem.MolToSmiles(frag)
                    fp_smiles[fragment_id] = {'smiles': smiles, 'count': count}
                fp = fp_smiles
        else:
            fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
                datapoint,
                self.radius,
                nBits=self.size,
                useChirality=self.chiral,
                useBondTypes=self.bonds,
                useFeatures=self.features)
            fp = np.asarray(fp, dtype=float)
        return fp

    def _get_column_names(self, **kwargs) -> list:
        descriptors = []
        for i in range(self.size):
            descriptors.append("f" + str(i))
        return descriptors

    def __hash__(self):
        return hash((self.radius, self.size, self.chiral, self.bonds, self.features,
                     self.sparse, self.smiles))

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return self.radius == other.radius and \
               self.size == other.size and \
               self.chiral == other.chiral and \
               self.bonds == other.bonds and \
               self.features == other.features and \
               self.sparse == other.sparse and \
               self.smiles == other.smiles

    def _featurize_dataframe(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate circular fingerprint.
    Parameters
    ----------
    datapoint: rdkit.Chem.rdchem.Mol
      RDKit Mol object
    Returns
    -------
    np.ndarray
      A numpy array of circular fingerprint.
    """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdMolDescriptors
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        if self.sparse:
            info: Dict = {}
            fp = rdMolDescriptors.GetMorganFingerprint(
                datapoint,
                self.radius,
                useChirality=self.chiral,
                useBondTypes=self.bonds,
                useFeatures=self.features,
                bitInfo=info)
            fp = fp.GetNonzeroElements()  # convert to a dict

            # generate SMILES for fragments
            if self.smiles:
                fp_smiles = {}
                for fragment_id, count in fp.items():
                    root, radius = info[fragment_id][0]
                    env = Chem.FindAtomEnvironmentOfRadiusN(datapoint, radius, root)
                    frag = Chem.PathToSubmol(datapoint, env)
                    smiles = Chem.MolToSmiles(frag)
                    fp_smiles[fragment_id] = {'smiles': smiles, 'count': count}
                fp = fp_smiles
        else:
            fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
                datapoint,
                self.radius,
                nBits=self.size,
                useChirality=self.chiral,
                useBondTypes=self.bonds,
                useFeatures=self.features)
            fp = np.asarray(fp, dtype=float)
        return fp



from __future__ import annotations

from abc import abstractmethod
from typing import Generic, Optional, SupportsFloat, Tuple, TypeVar, Union
from jaqpotpy.utils import seeding
from jaqpotpy.rl import spaces

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class MolEnv(Generic[ObsType, ActType]):
    """The main Jaqpotpy Gym class. .
    The main API methods that users of this class need to know are:
        step
        reset
        render
        close
        seed
    And set the following attributes:
        action_space: The Space object corresponding to valid actions
        observation_space: The Space object corresponding to valid observations
        reward_range: A tuple corresponding to the min and max possible rewards
    Note: a default reward range set to [-inf,+inf] already exists. Set it if you want a narrower range.
    The methods are accessed publicly as "step", "reset", etc...
    """

    # Set this in SOME subclasses
    metadata = {"render_modes": []}
    reward_range = (-float("inf"), float("inf"))
    spec = None

    # Set these in ALL subclasses
    action_space: spaces.Space[ActType]
    observation_space: spaces.Space[ObsType]

    # Created
    _np_random: seeding.RandomNumberGenerator | None = None

    @property
    def np_random(self) -> seeding.RandomNumberGenerator:
        """Initializes the np_random field if not done already."""
        if self._np_random is None:
            self._np_random, seed = seeding.np_random()
        return self._np_random

    @np_random.setter
    def np_random(self, value: seeding.RandomNumberGenerator):
        self._np_random = value

    @abstractmethod
    def step(self, action: ActType) -> Tuple[ObsType, float, bool, dict]:
        """Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling :meth:`reset`
        to reset this environment's state.
        Accepts an action and returns a tuple (observation, reward, done, info).
        Args:
            action (object): an action provided by the agent
        This method returns a tuple ``(observation, reward, done, info)``
        Returns:
            observation (object): agent's observation of the current environment. This will be an element of the environment's :attr:`observation_space`. This may, for instance, be a numpy array containing the positions and velocities of certain objects.
            reward (float) : amount of reward returned after previous action
            done (bool): whether the episode has ended, in which case further :meth:`step` calls will return undefined results. A done signal may be emitted for different reasons: Maybe the task underlying the environment was solved successfully, a certain timelimit was exceeded, or the physics simulation has entered an invalid state. ``info`` may contain additional information regarding the reason for a ``done`` signal.
            info (dict): contains auxiliary diagnostic information (helpful for debugging, learning, and logging). This might, for instance, contain:

                - metrics that describe the agent's performance or
                - state variables that are hidden from observations or
                - information that distinguishes truncation and termination or
                - individual reward terms that are combined to produce the total reward
        """
        raise NotImplementedError

    @abstractmethod
    def reset(
            self,
            *,
            seed: Optional[int] = None,
            return_info: bool = False,
            options: Optional[dict] = None,
    ) -> Union[ObsType, tuple[ObsType, dict]]:
        """Resets the environment to an initial state and returns an initial
        observation.
        This method should also reset the environment's random number
        generator(s) if ``seed`` is an integer or if the environment has not
        yet initialized a random number generator. If the environment already
        has a random number generator and :meth:`reset` is called with ``seed=None``,
        the RNG should not be reset.
        Moreover, :meth:`reset` should (in the typical use case) be called with an
        integer seed right after initialization and then never again.
        Args:
            seed (int or None): The seed that is used to initialize the environment's PRNG. If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed, a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom). However, if the environment already has a PRNG and ``seed=None`` is pased, the PRNG will *not* be reset. If you pass an integer, the PRNG will be reset even if it already exists. Usually, you want to pass an integer *right after the environment has been initialized and then never again*. Please refer to the minimal example above to see this paradigm in action.
            return_info (bool): If true, return additional information along with initial observation. This info should be analogous to the info returned in :meth:`step`
            options (dict or None): Additional information to specify how the environment is reset (optional, depending on the specific environment)
        Returns:
            observation (object): Observation of the initial state. This will be an element of :attr:`observation_space` (usually a numpy array) and is analogous to the observation returned by :meth:`step`.
            info (optional dictionary): This will *only* be returned if ``return_info=True`` is passed. It contains auxiliary information complementing ``observation``. This dictionary should be analogous to the ``info`` returned by :meth:`step`.
        """
        # Initialize the RNG if the seed is manually passed
        if seed is not None:
            self._np_random, seed = seeding.np_random(seed)

    @abstractmethod
    def render(self, mode="human"):
        """Renders the environment.
        The set of supported modes varies per environment. (And some
        third-party environments may not support rendering at all.)
        By convention, if mode is:
        - human: render to the current display or terminal and
          return nothing. Usually for human consumption.
        - rgb_array: Return an numpy.ndarray with shape (x, y, 3),
          representing RGB values for an x-by-y pixel image, suitable
          for turning into a video.
        - ansi: Return a string (str) or StringIO.StringIO containing a
          terminal-style text representation. The text can include newlines
          and ANSI escape sequences (e.g. for colors).

        Note:
            Make sure that your class's metadata 'render_modes' key includes
              the list of supported modes. It's recommended to call super()
              in implementations to use the functionality of this method.
        Args:
            mode (str): the mode to render with
        Example::
            class MyEnv(Env):
                metadata = {'render_modes': ['human', 'rgb_array']}
                def render(self, mode='human'):
                    if mode == 'rgb_array':
                        return np.array(...) # return RGB frame suitable for video
                    elif mode == 'human':
                        ... # pop up a window and render
                    else:
                        super(MyEnv, self).render(mode=mode) # just raise an exception
        """
        raise NotImplementedError

    def close(self):
        """Override close in your subclass to perform any necessary cleanup.
        Environments will automatically close() themselves when
        garbage collected or when the program exits.
        """
        pass

    def seed(self, seed=None):
        """Sets the seed for this env's random number generator(s).
        Note:
            Some environments use multiple pseudorandom number generators.
            We want to capture all such seeds used in order to ensure that
            there aren't accidental correlations between multiple generators.
        Returns:
            list<bigint>: Returns the list of seeds used in this env's random
              number generators. The first value in the list should be the
              "main" seed, or the value which a reproducer should pass to
              'seed'. Often, the main seed equals the provided 'seed', but
              this won't be true if seed=None, for example.
        """
        self._np_random, seed = seeding.np_random(seed)
        return [seed]

    @property
    def unwrapped(self) -> Env:
        """Completely unwrap this env.
        Returns:
            gym.Env: The base non-wrapped gym.Env instance
        """
        return self

    def __str__(self):
        if self.spec is None:
            return f"<{type(self).__name__} instance>"
        else:
            return f"<{type(self).__name__}<{self.spec.id}>>"

    def __enter__(self):
        """Support with-statement for the environment."""
        return self

    def __exit__(self, *args):
        """Support with-statement for the environment."""
        self.close()
        # propagate exception
        return False


class Env(Generic[ObsType, ActType]):
    """The main OpenAI Gym class. It encapsulates an environment with
    arbitrary behind-the-scenes dynamics. An environment can be
    partially or fully observed.
    The main API methods that users of this class need to know are:
        step
        reset
        render
        close
        seed
    And set the following attributes:
        action_space: The Space object corresponding to valid actions
        observation_space: The Space object corresponding to valid observations
        reward_range: A tuple corresponding to the min and max possible rewards
    Note: a default reward range set to [-inf,+inf] already exists. Set it if you want a narrower range.
    The methods are accessed publicly as "step", "reset", etc...
    """

    # Set this in SOME subclasses
    metadata = {"render_modes": []}
    reward_range = (-float("inf"), float("inf"))
    spec = None

    # Set these in ALL subclasses
    action_space: spaces.Space[ActType]
    observation_space: spaces.Space[ObsType]

    # Created
    _np_random: seeding.RandomNumberGenerator | None = None

    @property
    def np_random(self) -> seeding.RandomNumberGenerator:
        """Initializes the np_random field if not done already."""
        if self._np_random is None:
            self._np_random, seed = seeding.np_random()
        return self._np_random

    @np_random.setter
    def np_random(self, value: seeding.RandomNumberGenerator):
        self._np_random = value

    @abstractmethod
    def step(self, action: ActType) -> Tuple[ObsType, float, bool, dict]:
        """Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling :meth:`reset`
        to reset this environment's state.
        Accepts an action and returns a tuple (observation, reward, done, info).
        Args:
            action (object): an action provided by the agent
        This method returns a tuple ``(observation, reward, done, info)``
        Returns:
            observation (object): agent's observation of the current environment. This will be an element of the environment's :attr:`observation_space`. This may, for instance, be a numpy array containing the positions and velocities of certain objects.
            reward (float) : amount of reward returned after previous action
            done (bool): whether the episode has ended, in which case further :meth:`step` calls will return undefined results. A done signal may be emitted for different reasons: Maybe the task underlying the environment was solved successfully, a certain timelimit was exceeded, or the physics simulation has entered an invalid state. ``info`` may contain additional information regarding the reason for a ``done`` signal.
            info (dict): contains auxiliary diagnostic information (helpful for debugging, learning, and logging). This might, for instance, contain:

                - metrics that describe the agent's performance or
                - state variables that are hidden from observations or
                - information that distinguishes truncation and termination or
                - individual reward terms that are combined to produce the total reward
        """
        raise NotImplementedError

    @abstractmethod
    def reset(
            self,
            *,
            seed: Optional[int] = None,
            return_info: bool = False,
            options: Optional[dict] = None,
    ) -> Union[ObsType, tuple[ObsType, dict]]:
        """Resets the environment to an initial state and returns an initial
        observation.
        This method should also reset the environment's random number
        generator(s) if ``seed`` is an integer or if the environment has not
        yet initialized a random number generator. If the environment already
        has a random number generator and :meth:`reset` is called with ``seed=None``,
        the RNG should not be reset.
        Moreover, :meth:`reset` should (in the typical use case) be called with an
        integer seed right after initialization and then never again.
        Args:
            seed (int or None): The seed that is used to initialize the environment's PRNG. If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed, a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom). However, if the environment already has a PRNG and ``seed=None`` is pased, the PRNG will *not* be reset. If you pass an integer, the PRNG will be reset even if it already exists. Usually, you want to pass an integer *right after the environment has been initialized and then never again*. Please refer to the minimal example above to see this paradigm in action.
            return_info (bool): If true, return additional information along with initial observation. This info should be analogous to the info returned in :meth:`step`
            options (dict or None): Additional information to specify how the environment is reset (optional, depending on the specific environment)
        Returns:
            observation (object): Observation of the initial state. This will be an element of :attr:`observation_space` (usually a numpy array) and is analogous to the observation returned by :meth:`step`.
            info (optional dictionary): This will *only* be returned if ``return_info=True`` is passed. It contains auxiliary information complementing ``observation``. This dictionary should be analogous to the ``info`` returned by :meth:`step`.
        """
        # Initialize the RNG if the seed is manually passed
        if seed is not None:
            self._np_random, seed = seeding.np_random(seed)

    @abstractmethod
    def render(self, mode="human"):
        """Renders the environment.
        The set of supported modes varies per environment. (And some
        third-party environments may not support rendering at all.)
        By convention, if mode is:
        - human: render to the current display or terminal and
          return nothing. Usually for human consumption.
        - rgb_array: Return an numpy.ndarray with shape (x, y, 3),
          representing RGB values for an x-by-y pixel image, suitable
          for turning into a video.
        - ansi: Return a string (str) or StringIO.StringIO containing a
          terminal-style text representation. The text can include newlines
          and ANSI escape sequences (e.g. for colors).

        Note:
            Make sure that your class's metadata 'render_modes' key includes
              the list of supported modes. It's recommended to call super()
              in implementations to use the functionality of this method.
        Args:
            mode (str): the mode to render with
        Example::
            class MyEnv(Env):
                metadata = {'render_modes': ['human', 'rgb_array']}
                def render(self, mode='human'):
                    if mode == 'rgb_array':
                        return np.array(...) # return RGB frame suitable for video
                    elif mode == 'human':
                        ... # pop up a window and render
                    else:
                        super(MyEnv, self).render(mode=mode) # just raise an exception
        """
        raise NotImplementedError

    def close(self):
        """Override close in your subclass to perform any necessary cleanup.
        Environments will automatically close() themselves when
        garbage collected or when the program exits.
        """
        pass

    def seed(self, seed=None):
        """Sets the seed for this env's random number generator(s).
        Note:
            Some environments use multiple pseudorandom number generators.
            We want to capture all such seeds used in order to ensure that
            there aren't accidental correlations between multiple generators.
        Returns:
            list<bigint>: Returns the list of seeds used in this env's random
              number generators. The first value in the list should be the
              "main" seed, or the value which a reproducer should pass to
              'seed'. Often, the main seed equals the provided 'seed', but
              this won't be true if seed=None, for example.
        """
        self._np_random, seed = seeding.np_random(seed)
        return [seed]

    @property
    def unwrapped(self) -> Env:
        """Completely unwrap this env.
        Returns:
            gym.Env: The base non-wrapped gym.Env instance
        """
        return self

    def __str__(self):
        if self.spec is None:
            return f"<{type(self).__name__} instance>"
        else:
            return f"<{type(self).__name__}<{self.spec.id}>>"

    def __enter__(self):
        """Support with-statement for the environment."""
        return self

    def __exit__(self, *args):
        """Support with-statement for the environment."""
        self.close()
        # propagate exception
        return False


class Wrapper(Env[ObsType, ActType]):
    """Wraps the environment to allow a modular transformation.
    This class is the base class for all wrappers. The subclass could override
    some methods to change the behavior of the original environment without touching the
    original code.
    .. note::
        Don't forget to call ``super().__init__(env)`` if the subclass overrides :meth:`__init__`.
    """

    def __init__(self, env: Env):
        self.env = env

        self._action_space: spaces.Space | None = None
        self._observation_space: spaces.Space | None = None
        self._reward_range: tuple[SupportsFloat, SupportsFloat] | None = None
        self._metadata: dict | None = None

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(f"accessing private attribute '{name}' is prohibited")
        return getattr(self.env, name)

    @property
    def spec(self):
        return self.env.spec

    @classmethod
    def class_name(cls):
        return cls.__name__

    @property
    def action_space(self) -> spaces.Space[ActType]:
        if self._action_space is None:
            return self.env.action_space
        return self._action_space

    @action_space.setter
    def action_space(self, space):
        self._action_space = space

    @property
    def observation_space(self) -> spaces.Space:
        if self._observation_space is None:
            return self.env.observation_space
        return self._observation_space

    @observation_space.setter
    def observation_space(self, space):
        self._observation_space = space

    @property
    def reward_range(self) -> tuple[SupportsFloat, SupportsFloat]:
        if self._reward_range is None:
            return self.env.reward_range
        return self._reward_range

    @reward_range.setter
    def reward_range(self, value):
        self._reward_range = value

    @property
    def metadata(self) -> dict:
        if self._metadata is None:
            return self.env.metadata
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, dict]:
        return self.env.step(action)

    def reset(self, **kwargs) -> Union[ObsType, tuple[ObsType, dict]]:
        return self.env.reset(**kwargs)

    def render(self, **kwargs):
        return self.env.render(**kwargs)

    def close(self):
        return self.env.close()

    def seed(self, seed=None):
        return self.env.seed(seed)

    def __str__(self):
        return f"<{type(self).__name__}{self.env}>"

    def __repr__(self):
        return str(self)

    @property
    def unwrapped(self) -> Env:
        return self.env.unwrapped


class ObservationWrapper(Wrapper):
    def reset(self, **kwargs):
        if kwargs.get("return_info", False):
            obs, info = self.env.reset(**kwargs)
            return self.observation(obs), info
        else:
            return self.observation(self.env.reset(**kwargs))

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        return self.observation(observation), reward, done, info

    @abstractmethod
    def observation(self, observation):
        raise NotImplementedError


class RewardWrapper(Wrapper):
    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        return observation, self.reward(reward), done, info

    @abstractmethod
    def reward(self, reward):
        raise NotImplementedError


class ActionWrapper(Wrapper):
    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

    def step(self, action):
        return self.env.step(self.action(action))

    @abstractmethod
    def action(self, action):
        raise NotImplementedError

    @abstractmethod
    def reverse_action(self, action):
        raise NotImplementedError


from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from jaqpotpy.descriptors.graph.graph_data import GraphData
from pymatgen.core.structure import Structure
import numpy as np
import json
from typing import Union, Tuple
import pandas as pd


class CrystalGraphCNN(MaterialFeaturizer):
  """
  Calculate structure graph features for crystals.
  Based on the implementation in Crystal Graph Convolutional
  Neural Networks (CGCNN). The method constructs a crystal graph
  representation including atom features and bond features (neighbor
  distances). Neighbors are determined by searching in a sphere around
  atoms in the unit cell. A Gaussian filter is applied to neighbor distances.
  All units are in angstrom.
  This featurizer requires the dependency pymatgen. It may
  be useful when 3D coordinates are available and when using graph
  network models and crystal graph convolutional networks.
  See [1]_ for more details.
  References
  ----------
  .. [1] T. Xie and J. C. Grossman, "Crystal graph convolutional
     neural networks for an accurate and interpretable prediction
     of material properties", Phys. Rev. Lett.
  Examples
  --------
  >>> import jaqpotpy as jt
  >>> from pymatgen.core.lattice import Lattice
  >>> from pymatgen.core.structure import Structure
  >>> lattice = Lattice.cubic(4.2)
  >>> struct = Structure(lattice, ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
  >>> featurizer = jt.descriptors.material.CrystalGraphCNN()
  >>> features = featurizer.featurize([struct])
  >>> type(features[0])
  <class 'jaqpotpy.descriptors.material.GraphData'>

  Note
  ----
  This class requires matminer and Pymatgen to be installed.
  `NaN` feature values are automatically converted to 0 by this featurizer.
  """

  def __init__(self, radius: float = 8.0, max_neighbors: int = 12, step:float = 0.2):
    """
     Parameters
    ----------
    radius: float (default 8.0)
      Radius of sphere for finding neighbors of atoms in unit cell.
    max_neighbors: int (default 12)
      Maximum number of neighbors to consider when constructing graph.
    step: float (default 0.2)
      Step size for Gaussian filter. This value is used when building edge features.
    """
    self.radius = radius
    self.max_neighbors = int(max_neighbors)
    self.step = step
    from jaqpotpy.descriptors.material.datafiles.atom_init import atom_init_json

    # print(atom_init_json)
    # with open('jaqpotpy/descriptors/material/datafiles/atom_init.json', 'r') as f:
    #     atom_init_json = json.load(f)

    self.atom_features = {
        int(key): np.array(value, dtype=np.float32)
        for key, value in atom_init_json.items()
    }
    self.valid_atom_number = set(self.atom_features.keys())

  @property
  def __name__(self):
    return 'CrystalGraphCNN'

  def __getitem__(self):
    return self


  def _featurize(self, datapoint: Union[str, Structure], **kwargs) -> GraphData:
    """
    Calculate crystal graph features from pymatgen structure.

    Parameters
    ----------
    datapoint: str or pymatgen.core.structure object
      Either the path of a file of a material (i.e. extxyz, cif etc.)
      from wich the structure will be created, or the Structure itself.
    Returns
    -------
    feats: np.ndarray
      Vector of properties and statistics derived from chemical
      stoichiometry. Some values may be NaN.
    """

    if isinstance(datapoint, str):
        if datapoint.split('.')[-1] == 'extxyz':
            from ase.io import read
            xyz = read(datapoint)
            lat = np.array(xyz.get_cell())
            sym = xyz.get_chemical_symbols()
            pos = xyz.arrays['positions']
            struct = Structure(lat, sym, pos)
        else:
            struct = Structure.from_file(datapoint)
    else:
        struct = datapoint

    node_features = self._get_node_features(struct)
    edge_index, edge_features = self._get_edge_features_and_index(struct)
    graph = GraphData(node_features, edge_index, edge_features)
    return graph


  def _featurize_dataframe(self, datapoint: Union[str, Structure], **kwargs) -> pd.DataFrame:
    """
    Calculate crystal graph features from pymatgen structure.

    Parameters
    ----------
    datapoint: str or pymatgen.core.structure object
      Either the path of a file of a material (i.e. extxyz, cif etc.)
      from wich the structure will be created, or the Structure itself.
    Returns
    -------
    feats: np.ndarray
      Vector of properties and statistics derived from chemical
      stoichiometry. Some values may be NaN.
    """

    df = pd.DataFrame(index=[0], columns=['MaterialGraph'])
    df.at[0, 'MaterialGraph'] = self._featurize(datapoint)
    return df

  def _get_node_features(self, struct: Structure) -> np.ndarray:
      """
      Get the node feature from `atom_init.json`. The `atom_init.json` was collected
      from `data/sample-regression/atom_init.json` in the CGCNN repository.
      Parameters
      ----------
      struct: pymatgen.core.Structure
        A periodic crystal composed of a lattice and a sequence of atomic
        sites with 3D coordinates and elements.
      Returns
      -------
      node_features: np.ndarray
        A numpy array of shape `(num_nodes, 92)`.
      """
      node_features = []
      for site in struct:
          # check whether the atom feature exists or not
          assert site.specie.number in self.valid_atom_number
          node_features.append(self.atom_features[site.specie.number])
      return np.vstack(node_features).astype(float)

  def _get_edge_features_and_index(self, struct: Structure) -> Tuple[np.ndarray, np.ndarray]:
      """
      Calculate the edge feature and edge index from pymatgen structure.
      Parameters
      ----------
      struct: pymatgen.core.Structure
        A periodic crystal composed of a lattice and a sequence of atomic
        sites with 3D coordinates and elements.
      Returns
      -------
      edge_idx np.ndarray, dtype int
        A numpy array of shape with `(2, num_edges)`.
      edge_features: np.ndarray
        A numpy array of shape with `(num_edges, filter_length)`. The `filter_length` is
        (self.radius / self.step) + 1. The edge features were built by applying gaussian
        filter to the distance between nodes.
      """

      neighbors = struct.get_all_neighbors(self.radius, include_index=True)
      neighbors = [sorted(n, key=lambda x: x[1]) for n in neighbors]

      # construct bi-directed graph
      src_idx, dest_idx = [], []
      edge_distances = []
      for node_idx, neighbor in enumerate(neighbors):
          neighbor = neighbor[:self.max_neighbors]
          src_idx.extend([node_idx] * len(neighbor))
          dest_idx.extend([site[2] for site in neighbor])
          edge_distances.extend([site[1] for site in neighbor])

      edge_idx = np.array([src_idx, dest_idx], dtype=int)
      edge_features = self._gaussian_filter(np.array(edge_distances, dtype=float))
      return edge_idx, edge_features

  def _gaussian_filter(self, distances: np.ndarray) -> np.ndarray:
      """
      Apply Gaussian filter to an array of interatomic distances.
      Parameters
      ----------
      distances : np.ndarray
        A numpy array of the shape `(num_edges, )`.
      Returns
      -------
      expanded_distances: np.ndarray
        Expanded distance tensor after Gaussian filtering.
        The shape is `(num_edges, filter_length)`. The `filter_length` is
        (self.radius / self.step) + 1.
      """

      filt = np.arange(0, self.radius + self.step, self.step)

      # Increase dimension of distance tensor and apply filter
      expanded_distances = np.exp(-(distances[..., np.newaxis] - filt) ** 2 / self.step ** 2)

      return expanded_distances




import torch
import torch.nn as nn
from torch_geometric.nn import GCNConv
from torch_geometric.nn import global_mean_pool
import numpy as np
from jaqpotpy.models.generative.models.layers import GraphConvolution, GraphAggregation\
    , GraphConvolutionLayer, GraphConvolution2
from typing import Any, Iterable, Union, Dict


class GanGeneratorModel(object):
    _model: Any

    def sample(self, batch_size):
        raise NotImplemented("Not implemented")

    @property
    def descriptors(self):
        return self._descriptors

    @descriptors.setter
    def descriptors(self, value):
        self._descriptors = value

    @property
    def preprocessing(self):
        return self._preprocessors

    @preprocessing.setter
    def preprocessing(self, value):
        self._preprocessors = value

    @property
    def preprocessing_y(self):
        return self._preprocessors_y

    @preprocessing_y.setter
    def preprocessing_y(self, value):
        self._preprocessors_y = value

    @property
    def preprocessor_names(self):
        return self._preprocessor_names

    @preprocessor_names.setter
    def preprocessor_names(self, value):
        self._preprocessor_names = value

    @property
    def preprocessor_y_names(self):
        return self._preprocessor_names

    @preprocessor_y_names.setter
    def preprocessor_y_names(self, value):
        self._preprocessor_y_names = value

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        self._X = value

    @property
    def Y(self):
        return self._Y

    @Y.setter
    def Y(self, value):
        self._Y = value

    @property
    def external(self) -> dict:
        return self._external

    @external.setter
    def external(self, value):
        self._external = value

    @property
    def external_feats(self) -> Iterable[str]:
        return self._external_feats

    @external_feats.setter
    def external_feats(self, value):
        self._external_feats = value

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def model_title(self):
        return self._model_title

    @model_title.setter
    def model_title(self, value):
        self._model_title = value

    @property
    def prediction(self):
        return self._prediction

    @prediction.setter
    def prediction(self, value):
        self._prediction = value

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, value):
        self._probability = value

    @property
    def library(self):
        return self._library

    @library.setter
    def library(self, value):
        self._library = value

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, value):
        self._optimizer = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def jaqpotpy_version(self):
        return self._jaqpotpy_version

    @jaqpotpy_version.setter
    def jaqpotpy_version(self, value):
        self._jaqpotpy_version = value

    @property
    def jaqpotpy_docker(self):
        return self._jaqpotpy_docker

    @jaqpotpy_docker.setter
    def jaqpotpy_docker(self, value):
        self._jaqpotpy_docker = value

    def __train__(self):
        raise NotImplemented("Not implemented")

    def __eval__(self):
        raise NotImplemented("Not implemented")

    def infer(self):
        raise NotImplemented("Not implemented")

class GanMoleculeGenerator(nn.Module, GanGeneratorModel):
    """Generator network."""
    def __init__(self, conv_dims, z_dim, vertexes, edges, nodes, dropout_rate):
        super(GanMoleculeGenerator, self).__init__()
        self.activation_f = torch.nn.Tanh()
        self.multi_dense_layer = MultiDenseLayer(z_dim, conv_dims, self.activation_f)

        self.vertexes = vertexes
        self.edges = edges
        self.nodes = nodes

        self.edges_layer = nn.Linear(conv_dims[-1], edges * vertexes * vertexes)
        self.nodes_layer = nn.Linear(conv_dims[-1], vertexes * nodes)

        self.dropoout = nn.Dropout(p=dropout_rate)

    def forward(self, x):
        output = self.multi_dense_layer(x)
        edges_logits = self.edges_layer(output).view(-1, self.edges, self.vertexes, self.vertexes)
        edges_logits = (edges_logits + edges_logits.permute(0, 1, 3, 2)) / 2
        edges_logits = self.dropoout(edges_logits.permute(0, 2, 3, 1))

        nodes_logits = self.nodes_layer(output)
        nodes_logits = self.dropoout(nodes_logits.view(-1, self.vertexes, self.nodes))

        return edges_logits, nodes_logits

    def sample_generator(self, batch_size):
        return torch.Tensor(np.random.normal(0, 1, size=(batch_size, self.z_dim)))

    def sample(self, batch_size):
        return torch.Tensor(np.random.normal(0, 1, size=(batch_size, self.z_dim)))


class _GanMoleculeGenerator(nn.Module, GanGeneratorModel):
    """Generator network."""
    def __init__(self, conv_dims, z_dim, vertexes, edges, nodes, dropout):
        super(_GanMoleculeGenerator, self).__init__()

        self.z_dim = z_dim
        self.vertexes = vertexes
        self.edges = edges
        self.nodes = nodes

        layers = []
        for c0, c1 in zip([z_dim]+conv_dims[:-1], conv_dims):
            layers.append(nn.Linear(c0, c1))
            layers.append(nn.Tanh())
            layers.append(nn.Dropout(p=dropout, inplace=True))
        self.layers = nn.Sequential(*layers)

        self.edges_layer = nn.Linear(conv_dims[-1], edges * vertexes * vertexes)
        self.nodes_layer = nn.Linear(conv_dims[-1], vertexes * nodes)
        self.dropoout = nn.Dropout(p=dropout)

    def forward(self, x):
        output = self.layers(x)
        edges_logits = self.edges_layer(output)\
                       .view(-1, self.edges, self.vertexes, self.vertexes)
        edges_logits = (edges_logits + edges_logits.permute(0, 1, 3, 2))/2
        edges_logits = self.dropoout(edges_logits.permute(0, 2, 3, 1))

        nodes_logits = self.nodes_layer(output)
        nodes_logits = self.dropoout(nodes_logits.view(-1,self.vertexes,self.nodes))

        return edges_logits, nodes_logits

    def sample_generator(self, batch_size):
        return torch.Tensor(np.random.normal(0, 1, size=(batch_size, self.z_dim)))

    def sample(self, batch_size):
        return torch.Tensor(np.random.normal(0, 1, size=(batch_size, self.z_dim)))


class GanMoleculeDiscriminator(nn.Module):
    """Discriminator network with PatchGAN."""
    def __init__(self, conv_dim, m_dim, b_dim, dropout):
        super(GanMoleculeDiscriminator, self).__init__()

        graph_conv_dim, aux_dim, linear_dim = conv_dim
        # discriminator
        self.gcn_layer = GCNConv(m_dim, graph_conv_dim, b_dim, dropout)
        self.agg_layer = global_mean_pool(graph_conv_dim[-1], aux_dim, b_dim, dropout)

        # multi dense layer
        layers = []
        for c0, c1 in zip([aux_dim]+linear_dim[:-1], linear_dim):
            layers.append(nn.Linear(c0, c1))
            layers.append(nn.Dropout(dropout))
        self.linear_layer = nn.Sequential(*layers)

        self.output_layer = nn.Linear(linear_dim[-1], 1)

    def forward(self, adj, hidden, node, activatation=None):
        adj = adj[:, :, :, 1:].permute(0, 3, 1, 2)
        annotations = torch.cat((hidden, node), -1) if hidden is not None else node
        h = self.gcn_layer(annotations, adj)
        annotations = torch.cat((h, hidden, node) if hidden is not None\
                                 else (h, node), -1)
        h = self.agg_layer(annotations, torch.tanh)
        h = self.linear_layer(h)

        # Need to implemente batch discriminator #
        ##########################################

        output = self.output_layer(h)
        output = activatation(output) if activatation is not None else output

        return output, h


class Discriminator2(nn.Module):
    """Discriminator network with PatchGAN."""

    def __init__(self, conv_dim, m_dim, b_dim, with_features=False, f_dim=0, dropout_rate=0.):
        super(Discriminator2, self).__init__()
        self.activation_f = torch.nn.Tanh()
        graph_conv_dim, aux_dim, linear_dim = conv_dim
        # discriminator
        self.gcn_layer = GraphConvolution2(m_dim, graph_conv_dim, b_dim, with_features, f_dim, dropout_rate)
        self.agg_layer = GraphAggregation(graph_conv_dim[-1] + m_dim, aux_dim, self.activation_f, with_features, f_dim,
                                          dropout_rate)
        self.multi_dense_layer = MultiDenseLayer(aux_dim, linear_dim, self.activation_f, dropout_rate=dropout_rate)

        self.output_layer = nn.Linear(linear_dim[-1], 1)

    def forward(self, adj, hidden, node, activation=None):
        adj = adj[:, :, :, 1:].permute(0, 3, 1, 2)
        h_1 = self.gcn_layer(node, adj, hidden)
        h = self.agg_layer(h_1, node, hidden)
        h = self.multi_dense_layer(h)

        output = self.output_layer(h)
        output = activation(output) if activation is not None else output

        return output, h


class MoleculeDiscriminator(nn.Module):
    """Discriminator network with PatchGAN."""

    def __init__(self, conv_dim, m_dim, b_dim, with_features=False, f_dim=0, dropout_rate=0.):
        super(MoleculeDiscriminator, self).__init__()
        self.activation_f = torch.nn.Tanh()
        graph_conv_dim, aux_dim, linear_dim = conv_dim
        # discriminator
        self.gcn_layer = GraphConvolution(m_dim, graph_conv_dim, b_dim, with_features, f_dim, dropout_rate)
        self.agg_layer = GraphAggregation(graph_conv_dim[-1] + m_dim, aux_dim, self.activation_f, with_features, f_dim,
                                          dropout_rate)
        self.multi_dense_layer = MultiDenseLayer(aux_dim, linear_dim, self.activation_f, dropout_rate=dropout_rate)

        self.output_layer = nn.Linear(linear_dim[-1], 1)

    def forward(self, adj, hidden, node, activation=None):
        adj = adj[:, :, :, 1:].permute(0, 3, 1, 2)
        h_1 = self.gcn_layer(node, adj, hidden)
        h = self.agg_layer(h_1, node, hidden)
        h = self.multi_dense_layer(h)
        output = self.output_layer(h)
        output = activation(output) if activation is not None else output
        return output, h


class MultiDenseLayer(nn.Module):
    def __init__(self, aux_unit, linear_units, activation=None, dropout_rate=0.):
        super(MultiDenseLayer, self).__init__()
        layers = []
        for c0, c1 in zip([aux_unit] + linear_units[:-1], linear_units):
            layers.append(nn.Linear(c0, c1))
            layers.append(nn.Dropout(dropout_rate))
            if activation is not None:
                layers.append(activation)
        self.linear_layer = nn.Sequential(*layers)

    def forward(self, inputs):
        h = self.linear_layer(inputs)
        return h


class Doa(object):

    def __init__(self, modelId=None, doaMatrix=None, aValue=None):
        self.modelId = None
        self.doaMatrix = None
        self.aValue = None

        if modelId is not None:
            self.modelId = modelId
        if doaMatrix is not None:
            self.doiMatrix = doaMatrix
        if aValue is not None:
            self.aValue = aValue



from pydantic import parse_obj_as
from typing import Union, List
from jaqpotpy.entities.chemical_models import ElementModel, IonicRadiiModel, IsotopesModel, \
    IonizationEnergiesModel, OxidationStatesModel, ScreeningConstantsModel
from jaqpotpy.helpers.periodic_table.base_classes import PeriodicTable


class Element(PeriodicTable):
    def __init__(self, element: Union[str, int]):
        super().__init__()
        if isinstance(element, str):
            if len(element) < 3:
                res = self._find_one('SELECT * FROM elements WHERE symbol="{}"'.format(element))
            else:
                res = self._find_one('SELECT * FROM elements WHERE name="{}"'.format(element))
        else:
            res = self._find_one('SELECT * FROM elements WHERE atomic_number="{}"'.format(element))

        try:
            g = self._find_one('SELECT name, symbol FROM groups WHERE group_id={}'.format(res['group_id']))
        except:
            g = {'name': None, 'symbol': None}
        try:
            s = self._find_one('SELECT name,color FROM series WHERE id={}'.format(res['series_id']))
        except:
            s = {'name': None, 'color': None}

        res['group_name'] = g['name']
        res['group_symbol'] = g['symbol']
        res['series_name'] = s['name']
        res['series_color'] = s['color']

        self._element = parse_obj_as(ElementModel, {k: None if not (v == v) else v for k, v in res.items()})

    @property
    def parameters(self) -> ElementModel:
        return self._element

    @property
    def ionic_radii(self) -> List[IonicRadiiModel]:
        res = self._find_many('SELECT * FROM ionicradii WHERE atomic_number={}'.format(self._element.atomic_number))
        return [parse_obj_as(IonicRadiiModel, {k: None if not (v == v) else v for k, v in rad.items()}) for rad in res]

    @property
    def ionization_energies(self) -> List[IonizationEnergiesModel]:
        res = self._find_many('SELECT * FROM ionizationenergies WHERE atomic_number={}'.format(self._element.atomic_number))
        return [parse_obj_as(IonizationEnergiesModel, {k: None if not (v == v) else v for k, v in rad.items()}) for rad in res]

    @property
    def isotopes(self) -> List[IsotopesModel]:
        res = self._find_many('SELECT * FROM isotopes WHERE atomic_number={}'.format(self._element.atomic_number))
        return [parse_obj_as(IsotopesModel, {k: None if not (v == v) else v for k, v in rad.items()}) for rad in res]

    @property
    def oxidation_states(self) -> List[OxidationStatesModel]:
        res = self._find_many('SELECT * FROM oxidationstates WHERE atomic_number={}'.format(self._element.atomic_number))
        return [parse_obj_as(OxidationStatesModel, {k: None if not (v == v) else v for k, v in rad.items()}) for rad in res]

    @property
    def screeningconstants(self) -> List[ScreeningConstantsModel]:
        res = self._find_many('SELECT * FROM screeningconstants WHERE atomic_number={}'.format(self._element.atomic_number))
        return [parse_obj_as(ScreeningConstantsModel, {k: None if not (v == v) else v for k, v in rad.items()}) for rad in res]



# from typing import Optional
# import numpy as np
#
#
# class GraphData:
#   """GraphData class
#   This data class is almost same as `torch_geometric.data.Data
#   <https://pytorch-geometric.readthedocs.io/en/latest/modules/data.html#torch_geometric.data.Data>`_.
#   Attributes
#   ----------
#   node_features: np.ndarray
#     Node feature matrix with shape [num_nodes, num_node_features]
#   edge_index: np.ndarray, dtype int
#     Graph connectivity in COO format with shape [2, num_edges]
#   edge_features: np.ndarray, optional (default None)
#     Edge feature matrix with shape [num_edges, num_edge_features]
#   node_pos_features: np.ndarray, optional (default None)
#     Node position matrix with shape [num_nodes, num_dimensions].
#   num_nodes: int
#     The number of nodes in the graph
#   num_node_features: int
#     The number of features per node in the graph
#   num_edges: int
#     The number of edges in the graph
#   num_edges_features: int, optional (default None)
#     The number of features per edge in the graph
#   Examples
#   --------
#   >>> import numpy as np
#   >>> node_features = np.random.rand(5, 10)
#   >>> edge_index = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]], dtype=np.int64)
#   >>> graph = GraphData(node_features=node_features, edge_index=edge_index)
#   """
#
#   def __init__(
#       self,
#       node_features: np.ndarray,
#       edge_index: np.ndarray,
#       edge_features: Optional[np.ndarray] = None,
#       node_pos_features: Optional[np.ndarray] = None,
#   ):
#     """
#     Parameters
#     ----------
#     node_features: np.ndarray
#       Node feature matrix with shape [num_nodes, num_node_features]
#     edge_index: np.ndarray, dtype int
#       Graph connectivity in COO format with shape [2, num_edges]
#     edge_features: np.ndarray, optional (default None)
#       Edge feature matrix with shape [num_edges, num_edge_features]
#     node_pos_features: np.ndarray, optional (default None)
#       Node position matrix with shape [num_nodes, num_dimensions].
#     """
#     # validate params
#     if isinstance(node_features, np.ndarray) is False:
#       raise ValueError('node_features must be np.ndarray.')
#
#     if isinstance(edge_index, np.ndarray) is False:
#       raise ValueError('edge_index must be np.ndarray.')
#     elif issubclass(edge_index.dtype.type, np.integer) is False:
#       raise ValueError('edge_index.dtype must contains integers.')
#     elif edge_index.shape[0] != 2:
#       raise ValueError('The shape of edge_index is [2, num_edges].')
#     elif np.max(edge_index) >= len(node_features):
#       raise ValueError('edge_index contains the invalid node number.')
#
#     if edge_features is not None:
#       if isinstance(edge_features, np.ndarray) is False:
#         raise ValueError('edge_features must be np.ndarray or None.')
#       elif edge_index.shape[1] != edge_features.shape[0]:
#         raise ValueError('The first dimension of edge_features must be the \
#                           same as the second dimension of edge_index.')
#
#     if node_pos_features is not None:
#       if isinstance(node_pos_features, np.ndarray) is False:
#         raise ValueError('node_pos_features must be np.ndarray or None.')
#       elif node_pos_features.shape[0] != node_features.shape[0]:
#         raise ValueError(
#             'The length of node_pos_features must be the same as the \
#                           length of node_features.')
#
#     self.node_features = node_features
#     self.edge_index = edge_index
#     self.edge_features = edge_features
#     self.node_pos_features = node_pos_features
#     self.num_nodes, self.num_node_features = self.node_features.shape
#     self.num_edges = edge_index.shape[1]
#     if self.edge_features is not None:
#       self.num_edge_features = self.edge_features.shape[1]
#
#   def to_pyg_graph(self):
#     """Convert to PyTorch Geometric graph data instance
#     Returns
#     -------
#     torch_geometric.data.Data
#       Graph data for PyTorch Geometric
#     Note
#     ----
#     This method requires PyTorch Geometric to be installed.
#     """
#     try:
#       import torch
#       from torch_geometric.data import Data
#     except ModuleNotFoundError:
#       raise ImportError(
#           "This function requires PyTorch Geometric to be installed.")
#
#     edge_features = self.edge_features
#     if edge_features is not None:
#       edge_features = torch.from_numpy(self.edge_features).float()
#     node_pos_features = self.node_pos_features
#     if node_pos_features is not None:
#       node_pos_features = torch.from_numpy(self.node_pos_features).float()
#
#     return Data(
#         x=torch.from_numpy(self.node_features).float(),
#         edge_index=torch.from_numpy(self.edge_index).long(),
#         edge_attr=edge_features,
#         pos=node_pos_features)
#
#   def to_dgl_graph(self, self_loop: bool = False):
#     """Convert to DGL graph data instance
#     Returns
#     -------
#     dgl.DGLGraph
#       Graph data for DGL
#     self_loop: bool
#       Whether to add self loops for the nodes, i.e. edges from nodes
#       to themselves. Default to False.
#     Note
#     ----
#     This method requires DGL to be installed.
#     """
#     try:
#       import dgl
#       import torch
#     except ModuleNotFoundError:
#       raise ImportError("This function requires DGL to be installed.")
#
#     src = self.edge_index[0]
#     dst = self.edge_index[1]
#
#     g = dgl.graph(
#         (torch.from_numpy(src).long(), torch.from_numpy(dst).long()),
#         num_nodes=self.num_nodes)
#     g.ndata['x'] = torch.from_numpy(self.node_features).float()
#
#     if self.node_pos_features is not None:
#       g.ndata['pos'] = torch.from_numpy(self.node_pos_features).float()
#
#     if self.edge_features is not None:
#       g.edata['edge_attr'] = torch.from_numpy(self.edge_features).float()
#
#     if self_loop:
#       # This assumes that the edge features for self loops are full-zero tensors
#       # In the future we may want to support featurization for self loops
#       g.add_edges(np.arange(self.num_nodes), np.arange(self.num_nodes))
#
#     return g
#



from pydantic import BaseModel
from typing import Optional, Union, List


class ElementModel(BaseModel):
    annotation: Optional[str]
    atomic_number: Optional[int]
    atomic_radius: Optional[float]
    atomic_volume: Optional[float]
    block: Optional[str]
    boiling_point: Optional[float]
    density: Optional[float]
    description: Optional[str]
    dipole_polarizability: Optional[float]
    electron_affinity: Optional[float]
    electronic_configuration: Optional[str]
    evaporation_heat: Optional[float]
    fusion_heat: Optional[float]
    group_name: Optional[str]
    group_symbol: Optional[str]
    lattice_constant: Optional[float]
    lattice_structure: Optional[str]
    melting_point: Optional[float]
    name: Optional[str]
    period: Optional[int]
    series_name: Optional[str]
    series_color: Optional[str]
    specific_heat: Optional[float]
    symbol: Optional[str]
    thermal_conductivity: Optional[float]
    vdw_radius: Optional[float]
    covalent_radius_cordero: Optional[float]
    covalent_radius_pyykko: Optional[float]
    en_pauling: Optional[float]
    en_allen: Optional[float]
    jmol_color: Optional[str]
    cpk_color: Optional[str]
    proton_affinity: Optional[float]
    gas_basicity: Optional[float]
    heat_of_formation: Optional[float]
    c6: Optional[float]
    covalent_radius_bragg: Optional[float]
    vdw_radius_bondi: Optional[float]
    vdw_radius_truhlar: Optional[float]
    vdw_radius_rt: Optional[float]
    vdw_radius_batsanov: Optional[float]
    vdw_radius_dreiding: Optional[float]
    vdw_radius_uff: Optional[float]
    vdw_radius_mm3: Optional[float]
    abundance_crust: Optional[float]
    abundance_sea: Optional[float]
    molcas_gv_color: Optional[str]
    en_ghosh: Optional[float]
    vdw_radius_alvarez: Optional[float]
    c6_gb: Optional[float]
    atomic_weight: Optional[float]
    atomic_weight_uncertainty: Optional[float]
    is_monoisotopic: Optional[bool]
    is_radioactive: Optional[bool]
    cas: Optional[str]
    atomic_radius_rahm: Optional[float]
    geochemical_class: Optional[str]
    goldschmidt_class: Optional[str]
    metallic_radius: Optional[float]
    metallic_radius_c12: Optional[float]
    covalent_radius_pyykko_double: Optional[str]
    covalent_radius_pyykko_triple: Optional[str]
    discoverers: Optional[str]
    discovery_year: Optional[int]
    discovery_location: Optional[str]
    name_origin: Optional[str]
    sources: Optional[str]
    uses: Optional[str]
    mendeleev_number: Optional[int]
    dipole_polarizability_unc: Optional[float]
    pettifor_number: Optional[int]
    glawe_number: Optional[int]


class IonicRadiiModel(BaseModel):
    atomic_number: int
    charge: int
    econf: Optional[str]
    coordination: Optional[str]
    spin: Optional[str]
    crystal_radius: Optional[float]
    ionic_radius: Optional[float]
    origin: Optional[str]
    most_reliable: Optional[float]


class IonizationEnergiesModel(BaseModel):
    atomic_number: int
    degree: int
    energy: Optional[float]


class IsotopesModel(BaseModel):
    atomic_number: int
    mass: Optional[float]
    abundance: Optional[float]
    mass_number: Optional[int]
    mass_uncertainty: Optional[float]
    is_radioactive: Optional[int]
    half_life: Optional[float]
    half_life_unit: Optional[str]
    spin: Optional[float]
    g_factor: Optional[float]
    quadrupole_moment: Optional[float]


class OxidationStatesModel(BaseModel):
    atomic_number: int
    oxidation_state: Optional[int]


class ScreeningConstantsModel(BaseModel):
    atomic_number: Optional[int]
    n: Optional[int]
    s: Optional[str]
    screening: Optional[float]





from sqlalchemy import create_engine
import pandas as pd
from typing import Union, List, Dict, Generator
from sqlalchemy import inspect


PATH = "jaqpotpy/helpers/periodic_table/elements.db"


class PeriodicTable(object):
    def __init__(self):
        self.__engine = create_engine("sqlite:///{path:s}".format(path=PATH), echo=False)

    @property
    def get_tables(self) -> List:
        inspector = inspect(self.__engine)
        return inspector.get_table_names()

    def _find_one(self, query) -> Dict:
        res = pd.read_sql_query(query, self.__engine)
        if len(res) > 1:
            raise Warning('Found more results for query. Returning the first one.')

        return res.loc[0].to_dict()

    def _find_many(self, query) -> Generator:
        res = pd.read_sql_query(query, self.__engine)

        for row in res.to_dict(orient='records'):
            yield row

    def _fetch_table(self, table, as_dataframe=False) -> Union[pd.DataFrame, Generator]:
        res = pd.read_sql_query('SELECT * FROM {}'.format(table), self.__engine)

        if as_dataframe:
            return res

        for row in res.to_dict(orient='records'):
            yield row


from jaqpotpy.descriptors.base_classes import MaterialFeaturizer
from pymatgen.util.typing import CompositionLike, Composition
import numpy as np
import pandas as pd

class ElementNet(MaterialFeaturizer):
  """
  Create a fixed vector of length 118, containing rwa, fractional
  element compostitions in a compound.

  References
  ----------
  Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
     https://doi.org/10.1038/s41598-018-35934-y

  Examples
  --------
  >>> import jaqpotpy as jt
  >>> comp = "Fe2O3"
  >>> featurizer = jt.descriptors.material.ElementNet()
  >>> features = featurizer.featurize([comp])
  >>> type(features[0])
  <class 'numpy.ndarray'>
  >>> features[0].shape
  (118,)
  Note
  ----
  This class requires matminer and Pymatgen to be installed.
  `NaN` feature values are automatically converted to 0 by this featurizer.
  """

  def __init__(self):
    """
   Create a fixed vector of length 118, containing rwa, fractional
    element compostitions in a compound.

      References
      ----------
      Jha, D., Ward, L., Paul, A. et al. Sci Rep 8, 17593 (2018).
         https://doi.org/10.1038/s41598-018-35934-y
    """
    self.elements = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
       'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr',
       'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br',
       'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
       'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
       'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
       'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
       'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
       'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
       'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',
       'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
    return

  @property
  def __name__(self):
    return 'ElementNet'

  def __getitem__(self):
    return self


  def _featurize(self, datapoint: CompositionLike, **kwargs) -> np.ndarray:
    """
    Calculate 118 dimensional vector containing fractional
    compositions of each element in the compound.

    Parameters
    ----------
    datapoint: pymatgen.utils.typing.CompositionLike object
      Composition or an object that can be used to creaet a composition

    Returns
    -------
    feats: np.ndarray
      Vector containing the fractional compositions of each element.
    """
    if not isinstance(datapoint, Composition):
        datapoint = Composition(datapoint)
    fractions = datapoint.fractional_composition.get_el_amt_dict()
    return np.array([fractions[e] if e in fractions else 0 for e in self.elements])

  def _featurize_dataframe(self, datapoint: CompositionLike, **kwargs) -> pd.DataFrame:
    """
    Calculate 118 dimensional vector containing fractional
    compositions of each element in the compound.

    Parameters
    ----------
    datapoint: pymatgen.utils.typing.CompositionLike object
      Composition or an object that can be used to create a composition

    Returns
    -------
    feats: pd.Dataframe
      Dataframe of the vector containing the fractional compositions of each element.
    """

    data = self._featurize(datapoint)
    return pd.DataFrame([data], columns=self.elements, index=[0])

